# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #

from typing import List

from parameterized import parameterized
from autoreduce_qp.systemtests.utils.data_archive import DataArchive

from autoreduce_webapp.selenium_tests.pages.configure_new_runs_page import ConfigureNewRunsPage
from autoreduce_webapp.selenium_tests.pages.variables_summary_page import VariableSummaryPage
from autoreduce_webapp.selenium_tests.tests.base_tests import (BaseTestCase, FooterTestMixin, NavbarTestMixin,
                                                               AccessibilityTestMixin)


class TestSeeInstrumentVariablesPageWithMissingFiles(BaseTestCase, NavbarTestMixin, FooterTestMixin,
                                                     AccessibilityTestMixin):
    """
    Test the SeeInstrumentVariablesPage without creating the test archive files first.
    This tests error cases, e.g. when reduce_vars is missing.
    """

    fixtures = BaseTestCase.fixtures + ["one_run_mixed_vars"]

    def setUp(self) -> None:
        """
        Sets up the VariableSummaryPage before each test
        """
        super().setUp()
        self.instrument_name = "TestInstrument"
        self.page = VariableSummaryPage(self.driver, self.instrument_name)
        self.page.launch()

    def test_edit_no_reduce_vars_shows_error(self):
        """Tests: Error is shown when no reduce_vars.py exists and edit variables is clicked."""
        self.page.click_run_edit_button_for(100100, 100150)
        message = self.page.message
        assert message.is_displayed()
        assert "No such file or directory" in message.text
        assert "user/scripts/autoreduction/reduce_vars.py" in message.text


class TestSeeInstrumentVariablesPage(BaseTestCase):
    fixtures = BaseTestCase.fixtures + ["one_run_mixed_vars"]

    @classmethod
    def setUpClass(cls):
        """Sets up the data archive with a reduce and reduce_vars script to be shared between test cases"""
        super().setUpClass()
        cls.instrument_name = "TestInstrument"
        cls.data_archive = DataArchive([cls.instrument_name], 21, 21)
        cls.data_archive.create()
        cls.data_archive.add_reduction_script(cls.instrument_name, """print('some text')""")
        # NOTE the value here must match the value of the last variable in the fixture
        # if it doesn't it's value will be updated once the previous range (100151-100199) is updated
        # this is expected, as the final variable (valid for 100200 onwards) should have the default script value!
        cls.data_archive.add_reduce_vars_script(cls.instrument_name, """standard_vars={"variable1":"value4"}""")

    @classmethod
    def tearDownClass(cls) -> None:
        """Destroys the created DataArchive"""
        cls.data_archive.delete()
        super().tearDownClass()

    def setUp(self) -> None:
        """Sets up the VariableSummaryPage before each test case"""
        super().setUp()
        self.page = VariableSummaryPage(self.driver, self.instrument_name)
        self.page.launch()

    @parameterized.expand([(100100, 100150, "value2", ["some new value", "value3", "value4"]),
                           (100151, 100199, "value3", ["value2", "some new value", "value4"]),
                           (100200, 0, "value4", ["value2", "value3", "some new value"])])
    def test_edit(self, start: int, end: int, value_to_modify: str, expected_strings: List[str]):
        """Test: That run variables can be edited for upcoming runs."""
        upcoming_panel = self.page.panels[1]
        # makes sure the value we are going to modify is present in the initial values
        assert value_to_modify in upcoming_panel.get_attribute("textContent")

        self.page.click_run_edit_button_for(start, end)

        new_runs_page = ConfigureNewRunsPage(self.driver, self.instrument_name, start, end)
        assert new_runs_page.run_start_val == str(start)

        assert new_runs_page.variable1_field_val == value_to_modify
        new_runs_page.variable1_field = "some new value"
        new_runs_page.submit_button.click()
        new_runs_page.replace_confirm.click()

        upcoming_panel = self.page.panels[1]
        # make sure the value we are modifying is no longer visible
        assert value_to_modify not in upcoming_panel.get_attribute("textContent")

        # checks that the newly edited value is present, but also that the unaffected ones are still there
        for expected in expected_strings:
            assert expected in upcoming_panel.get_attribute("textContent")

    @parameterized.expand([(100100, 100150, "value2", 100150), (100151, 100199, "value3", 100099),
                           (100200, 0, "value4", 100099)])
    def test_delete_variable_for_run(self, start: int, end: int, value_to_delete: str, end_run_for_current_vars: int):
        """Tests that variables can be deleted from upcoming runs"""
        upcoming_panel = self.page.panels[1]
        # makes sure the value we are going to modify is present in the initial values
        assert value_to_delete in upcoming_panel.get_attribute("textContent")

        incoming_run_numbers = upcoming_panel.find_elements_by_class_name("run-numbers")

        assert "100100" in incoming_run_numbers[0].text
        assert "100150" in incoming_run_numbers[0].text
        assert "100151" in incoming_run_numbers[1].text
        assert "100199" in incoming_run_numbers[1].text
        assert "100200" in incoming_run_numbers[2].text
        assert "Ongoing" in incoming_run_numbers[2].text

        self.page.click_run_delete_button_for(start, end)

        current_panel_runs = self.page.panels[0].find_element_by_class_name("run-numbers")
        # check that the current variables end at the correct run
        assert str(end_run_for_current_vars) in current_panel_runs.text
        upcoming_panel = self.page.panels[1]
        # make sure the value we are modifying is no longer visible
        assert value_to_delete not in upcoming_panel.get_attribute("textContent")

    @parameterized.expand([[1234567], [7654321]])
    def test_edit_experiment(self, experiment_reference):
        """Tests: Variables can be edited for upcoming runs for an experiment"""
        experiment_panel = self.page.panels[2]
        # makes sure the value we are going to modify is present in the initial values
        assert f"experiment {experiment_reference} var" in experiment_panel.get_attribute("textContent")

        self.page.click_experiment_edit_button_for(experiment_reference)
        new_runs_page = ConfigureNewRunsPage(self.driver,
                                             self.instrument_name,
                                             experiment_reference=experiment_reference)

        assert new_runs_page.variable1_field_val == f"experiment {experiment_reference} var"

        new_runs_page.variable1_field = "some new value"
        new_runs_page.submit_button.click()

        experiment_panel = self.page.panels[2]
        assert f"experiment {experiment_reference} var" not in experiment_panel.get_attribute("textContent")
        assert "some new value" in experiment_panel.get_attribute("textContent")

    @parameterized.expand([[1234567], [7654321]])
    def test_delete_variable_for_experiment(self, experiment_reference):
        """Tests: Variables can be deleted for an experiment"""
        experiment_panel = self.page.panels[2]
        # makes sure the value we are going to modify is present in the initial values
        assert f"experiment {experiment_reference} var" in experiment_panel.get_attribute("textContent")

        self.page.click_experiment_delete_button_for(experiment_reference)
        experiment_panel = self.page.panels[2]
        assert f"experiment {experiment_reference} var" not in experiment_panel.get_attribute("textContent")



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Selenium tests for the runs summary page
"""

from autoreduce_db.reduction_viewer.models import ReductionRun

from autoreduce_webapp.selenium_tests.pages.run_summary_page import RunSummaryPage
from autoreduce_webapp.selenium_tests.tests.base_tests import BaseTestCase, FooterTestMixin, NavbarTestMixin, \
    AccessibilityTestMixin


# pylint:disable=no-member
class TestRunSummaryPageNoArchive(NavbarTestMixin, BaseTestCase, FooterTestMixin, AccessibilityTestMixin):
    fixtures = BaseTestCase.fixtures + ["run_with_one_variable"]

    @classmethod
    def setUpClass(cls):
        """Set the instrument for all test cases"""
        super().setUpClass()
        cls.instrument_name = "TestInstrument"

    def setUp(self) -> None:
        """Set up RunSummaryPage before each test case"""
        super().setUp()
        self.page = RunSummaryPage(self.driver, self.instrument_name, 99999, 0)
        self.page.launch()

    def test_opening_run_summary_without_reduce_vars(self):
        """
        Test that opening the run summary without a reduce_vars present for the instrument
        will not show the "Reset to current" buttons as there is no current values!
        """
        # the reset to current values should not be visible
        assert self.page.warning_message.is_displayed()
        assert self.page.warning_message.text == ("The reduce_vars.py script is missing for this instrument."
                                                  " Please create it before being able to submit re-runs.")

    def test_opening_run_summary_without_run_variables(self):
        """
        Test that opening the run summary without a reduce_vars present for the instrument
        will not show the "Reset to current" buttons as there is no current values!
        """
        # Delete the variables, and re-open the page
        ReductionRun.objects.get(pk=1).run_variables.all().delete()
        self.page.launch()
        # the reset to current values should not be visible
        assert self.page.warning_message.is_displayed()
        assert self.page.warning_message.text == "No variables found for this run."
        assert self.page.run_description_text() == "Run description: This is the test run_description"



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Handle page responses for WebApp
"""
from django.http import HttpRequest
from django.shortcuts import render
from django.template import RequestContext

from autoreduce_db.reduction_viewer.models import Setting

# pylint: disable=unused-argument


def get_admin_email():
    """Check the settings for a valid admin email"""
    try:
        admin_email = Setting.objects.get(name='admin_email')
        return admin_email.value
    # pylint: disable=bare-except
    except:
        return ''


def handler400(request, exception):
    """Error 400 handler"""
    response = render(None, '400.html', {'admin_email': get_admin_email()}, RequestContext(request))
    response.status_code = 400
    return response


def handler404(request, exception):
    """Error 404 handler"""
    response = render('404.html', {'admin_email': get_admin_email()}, RequestContext(request))
    response.status_code = 404
    return response


def handler403(request, exception):
    """Error 403 handler"""
    response = render('403.html', {'admin_email': get_admin_email()}, RequestContext(request))
    response.status_code = 403
    return response


def handler500(request):
    """Error 500 handler"""
    response = render(RequestContext(request), '500.html', {'admin_email': get_admin_email()})
    response.status_code = 500
    return response


def render_error(request: HttpRequest, message: str):
    """
    Return the error page with a message displayed.
    :param request: (HttpRequest) The original sent request
    :param message: (str) The message that will be displayed
    :return: (HttpResponse) The error page
    """
    return render(request, 'error.html', {'message': message, 'admin_email': get_admin_email()}, status=500)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module containing configuration functions. Configuration file must be used as a thread safe way to
share values between xdist workers does not currently exist.
"""
import json
import sys
from pathlib import Path
from shutil import copyfile

from autoreduce_webapp.autoreduce_django.settings import CONFIG_ROOT

SELENIUM_CONFIG_DIR = Path(CONFIG_ROOT, "selenium_tests")
SELENIUM_CONFIG = Path(SELENIUM_CONFIG_DIR, "config.json")
TEMP_SELENIUM_CONFIG = Path(SELENIUM_CONFIG_DIR, "temp_config.json")

if not SELENIUM_CONFIG_DIR.exists():
    SELENIUM_CONFIG_DIR.mkdir(parents=True, exist_ok=True)

if not SELENIUM_CONFIG.exists():
    SELENIUM_CONFIG.write_text("""{"url": "http://localhost:0000","run_headless": true}""")


def store_original_config():
    """
    Make a copy of the config file to a temporary file. This prevents arguments given to the test
     entrypoint being persisted to the config file.
    """
    try:
        copyfile(SELENIUM_CONFIG, TEMP_SELENIUM_CONFIG)
    except OSError:
        sys.exit(f"Config file: {SELENIUM_CONFIG} could not be loaded...")


def get_url():
    """
    Returns the url to test against from the config
    :return: (str) The url to test against from the config
    """

    return load_config_file()["url"]


def is_headless():
    """
    Returns the headless boolean from the config
    :return: (bool) The headless boolean from the config
    """
    return load_config_file()["run_headless"]


def set_url(url):
    """
    Set the url to test against in the config. IPs must be prefixed with http/https still
    :param url: (str) The url to test against
    """
    config = load_config_file()
    config["url"] = url
    dump_to_config_file(config)


def set_headless(headless):
    """
    Set the headless option in the config to decide whether or not to use a headless driver
    :param headless: (bool) the headless bool option
    """
    config = load_config_file()
    config["run_headless"] = headless
    dump_to_config_file(config)


def cleanup_config():
    """
    Copy the original values back to the original config file, so as not to persist arguments given
    to test runner
    """
    copyfile(TEMP_SELENIUM_CONFIG, SELENIUM_CONFIG)
    if TEMP_SELENIUM_CONFIG.exists():
        TEMP_SELENIUM_CONFIG.unlink()


def load_config_file():
    """
    Load the config file into a python dictionary
    :return: (dict) The config file as a python dictionary
    """
    try:
        with open(SELENIUM_CONFIG) as fle:
            return json.load(fle)
    except FileNotFoundError:
        sys.exit(f"Config file is missing. Please create: {str(SELENIUM_CONFIG)}")


def dump_to_config_file(config_dict):
    """
    Dump the given dictionary to the config file
    :param config_dict: (dict) the dictionary to be dumped.
    """
    with open(SELENIUM_CONFIG, "w") as fle:
        json.dump(config_dict, fle, indent=4)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module containing the NavbarMixin
"""
from typing import List

from selenium.webdriver.remote.webelement import WebElement


class NavbarMixin:
    """
    NavbarMixin adds functionality of the navbar to a page when inherited
    """
    NAVBAR_CLASS = "navbar"
    LOGO_CLASS = "navbar-brand"
    LINKS_ID = "navbar_links"
    ALL_INSTRUMENTS_LINK_XPATH = ".//a[contains(text(), 'All Instruments')]"
    JOB_QUEUE_XPATH = ".//a[contains(text(), 'Job Queue')]"
    FAILED_JOBS_XPATH = ".//a[contains(text(), 'Failed Jobs')]"
    GRAPHS_XPATH = ".//a[contains(text(), 'Graphs')]"
    HELP_XPATH = ".//a[contains(text(), 'Help')]"

    def click_navbar_logo(self):
        """
        Click the brand logo in the navbar and return the current page
        """
        self.driver.find_element_by_class_name(self.NAVBAR_CLASS).click()
        return self

    def click_navbar_all_instruments(self):
        """
        Click the all instruments link in the navbar
        """
        self.driver.find_element_by_xpath(self.ALL_INSTRUMENTS_LINK_XPATH).click()

    def click_navbar_job_queue(self):
        """
        Click the job queue link in the navbar
        """
        self.driver.find_element_by_xpath(self.JOB_QUEUE_XPATH).click()

    def click_navbar_failed_jobs(self):
        """
        Click the failed jobs link in the navbar
        """
        self.driver.find_element_by_xpath(self.FAILED_JOBS_XPATH).click()

    def click_navbar_graphs(self):
        """
        Click the graphs link in the navbar
        """
        self.driver.find_element_by_xpath(self.GRAPHS_XPATH).click()

    def click_navbar_help(self):
        """
        Click the help link in the navbar
        """
        self.driver.find_element_by_xpath(self.HELP_XPATH).click()

    def _get_navbar(self) -> WebElement:
        return self.driver.find_element_by_class_name(self.NAVBAR_CLASS)

    def _get_logo(self) -> WebElement:
        return self.driver.find_element_by_class_name(self.LOGO_CLASS)

    def _get_navbar_links(self) -> WebElement:
        return self.driver.find_element_by_id(self.LINKS_ID)

    def is_navbar_logo_visible(self) -> bool:
        """
        Check if the brand logo is visible in the navbar
        :return: (bool) True if logo is visible, otherwise False
        """
        return self._get_logo().is_displayed()

    def is_navbar_visible(self) -> bool:
        """
        Check if the navbar is visible on a page
        :return: (bool) True if navbar is visible, otherwise False
        """
        return self._get_navbar().is_displayed()

    def is_navbar_links_visible(self) -> bool:
        """
        Check if the navbar links are visible
        :return: (bool) True if links are visible, otherwise False
        """
        return self._get_navbar_links().is_displayed()

    def get_notification_messages(self) -> List[str]:
        """
        Get the notification messages on the page.
        :return: (list of str) The text of the notifications messages
        """
        return [notification.text for notification in self.driver.find_elements_by_class_name("notification-message")]



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module for the run summary page model
"""
from django.urls.base import reverse
from selenium.webdriver.remote.webelement import WebElement
from autoreduce_webapp.selenium_tests.pages.component_mixins.footer_mixin import FooterMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.navbar_mixin import NavbarMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.rerun_form_mixin import RerunFormMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.tour_mixin import TourMixin
from autoreduce_webapp.selenium_tests.pages.page import Page


class RerunJobsPage(Page, RerunFormMixin, NavbarMixin, FooterMixin, TourMixin):
    """
    Page model class for run summary page
    """
    def __init__(self, driver, instrument):
        super().__init__(driver)
        self.instrument = instrument

    def url_path(self) -> str:
        """
        Return the current URL of the page.
        :return: (str) the url path
        """
        return reverse("instrument:submit_runs", kwargs={
            "instrument": self.instrument,
        })

    @property
    def form_validation_message(self) -> WebElement:
        """Finds and returns the form validation message"""
        return self.driver.find_element_by_id("form_validation_message")

    @property
    def form(self) -> WebElement:
        """Finds and returns the rerun form on the page"""
        return self.driver.find_element_by_id("submit_jobs")

    @property
    def run_range_field(self) -> WebElement:
        """
        Finds and returns the back button for toggling the form on the page.
        """
        return self.driver.find_element_by_id("run_range")

    @run_range_field.setter
    def run_range_field(self, value) -> None:
        """
        Finds and returns the back button for toggling the form on the page.
        """
        self._set_field(self.run_range_field, value)

    @property
    def reset_to_current_values(self) -> WebElement:
        """
        Finds and returns the "Reset to values in the current reduce_vars script" button
        """
        return self.driver.find_element_by_id("currentScript")

    @property
    def error_container(self) -> WebElement:
        """
        Returns the container of the error message
        """
        return self.driver.find_element_by_id("error_container")

    def error_message_text(self) -> str:
        """
        Returns the text shown in the error message
        """
        return self.driver.find_element_by_id("error_message").text



#  Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
#  Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
#  SPDX - License - Identifier: GPL-3.0-or-later
#
"""
Module to store autoreduce custom context processors.
"""


def support_email_processor(_):
    """
    Context processor to add support_email to every page for footer and other areas.
    """
    return {"support_email": "isisreduce@stfc.ac.uk"}



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module containing the footer mixin
"""


class FooterMixin:
    """
    Footer mixin contains functionality for the page footer when inherited
    """
    FOOTER_ID = "footer"
    HELP_LINK_ID = "footer-help"
    GITHUB_LINK_ID = "footer-github"
    SUPPORT_EMAIL_LINK_ID = "footer-email"

    def is_footer_visible(self):
        """
        Checks whether the footer is visible
        :returns: (bool) True if footer is visible, otherwise False
        """
        return self.driver.find_element_by_id(self.FOOTER_ID).is_displayed()

    def click_footer_help_link(self):
        """
        Clicks the help link in the footer
        """
        help_link = self.driver.find_element_by_id(self.HELP_LINK_ID)
        help_link.click()
        return self

    def click_footer_github_link(self):
        """
        Clicks the github link in the footer
        """
        github_link = self.driver.find_element_by_id(self.GITHUB_LINK_ID)
        github_link.click()
        return self

    def is_footer_email_visible(self):
        """
        Check if the support email is visible within the header
        :return: (bool) True if visible, otherwise False
        """
        return self.driver.find_element_by_id(self.SUPPORT_EMAIL_LINK_ID).is_displayed()



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module for the overview page model
"""
from django.urls.base import reverse

from selenium.webdriver.common.by import By
from selenium.webdriver.support.expected_conditions import presence_of_element_located
from selenium.webdriver.support.wait import WebDriverWait

from autoreduce_webapp.selenium_tests import configuration
from autoreduce_webapp.selenium_tests.pages.component_mixins.footer_mixin import FooterMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.navbar_mixin import NavbarMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.tour_mixin import TourMixin
from autoreduce_webapp.selenium_tests.pages.runs_list_page import RunsListPage
from autoreduce_webapp.selenium_tests.pages.page import Page


class OverviewPage(Page, NavbarMixin, FooterMixin, TourMixin):
    """
    Overview page model class
    """
    def __init__(self, driver):
        super().__init__(driver)
        self.step = 0

    def launch(self):
        """
        This is a bit of a hack to get around the fact that you will only be logged in when
        connecting to / and not /overview. Once we have a better way to simulate logging in on the
        dev/local environments we can remove this method and add proper logging in methods and tests
        """
        self.driver.get(configuration.get_url())
        self.driver.get(self.url())
        WebDriverWait(self.driver, 10).until(presence_of_element_located((By.CLASS_NAME, "instrument-btn")))
        return self

    @staticmethod
    def url_path():
        """
        Return the path section of the overview page url
        :return: (str) Path section of the page url
        """
        return reverse("overview")

    def _get_instrument_buttons(self):
        return self.driver.find_elements_by_class_name("instrument-btn")

    def get_instruments_from_buttons(self):
        """
        Gets the names of the instruments which have buttons on the overview page
        :return: (List) The instrument names which have buttons on the overview page
        """
        return [instrument_btn.get_attribute("id").split("-")[0] for instrument_btn in self._get_instrument_buttons()]

    def click_instrument(self, instrument: str) -> RunsListPage:
        """
        Clicks the instrument button for the given instrument
        :param instrument: (str) instrument name
        :return: (InstrumentSummaryPage) The overview page object
        """
        self.driver.find_element_by_id(f"{instrument}-instrument-btn").click()
        return RunsListPage(self.driver, instrument)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module for the instrument summary page model
"""
from typing import List

from django.urls import reverse
from selenium.common.exceptions import NoSuchElementException
from autoreduce_webapp.selenium_tests.pages.component_mixins.footer_mixin import FooterMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.navbar_mixin import NavbarMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.tour_mixin import TourMixin
from autoreduce_webapp.selenium_tests.pages.page import Page
from autoreduce_webapp.selenium_tests.pages.run_summary_page import RunSummaryPage


class RunsListPage(Page, NavbarMixin, FooterMixin, TourMixin):
    """
    Page model class for instrument summary page
    """
    def __init__(self, driver, instrument):
        super().__init__(driver)
        self.instrument = instrument

    def url_path(self):
        """
        Return the path section of the instrument url
        :return: (str) Path section of the page url
        """
        return reverse("runs:list", kwargs={"instrument": self.instrument})

    def get_run_numbers_from_table(self) -> List[str]:
        """
        Get the list of run numbers visible on the current table of the instrument summary page
        :return: (List) List of strings of the run numbers of the current instrument summary page
        """
        return [run.text.split(" - ")[0] for run in self.driver.find_elements_by_class_name("run-num-links")]

    def click_run(self, run_number: int, version: int = 0) -> RunSummaryPage:
        """
        Click the run number link on the instrument summary table matching the given run number and
        version
        :param run_number: (int) the run number of the link to click
        :param version: (int) the version of the run to click
        :return: (RunSummaryPage) The page object of the opened run summary page.
        """
        runs = self.driver.find_elements_by_class_name("run-num-links")
        run_string = f"{run_number} - {version}" if version else f"{run_number}"
        for run in runs:
            if run.text == run_string:
                run.click()
                return RunSummaryPage(self.driver, self.instrument, run_number, version)
        raise NoSuchElementException

    def alert_message_text(self) -> str:
        """Get the the from the alert message"""
        return self.driver.find_element_by_id("alert_message").text.strip()



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #

import tempfile

from autoreduce_webapp.selenium_tests.pages.rerun_jobs_page import RerunJobsPage
from autoreduce_webapp.selenium_tests.tests.base_tests import (BaseTestCase, FooterTestMixin, NavbarTestMixin,
                                                               AccessibilityTestMixin)
from autoreduce_webapp.selenium_tests.utils import submit_and_wait_for_result

from autoreduce_webapp.selenium_tests.utils import setup_external_services

TEMP_OUT_FILE = tempfile.NamedTemporaryFile()  # pylint:disable=consider-using-with
SCRIPT = f"""
import sys
import os

sys.path.append(os.path.dirname(__file__))

import reduce_vars as web_var

def main(input_file, output_dir):
    with open("{TEMP_OUT_FILE.name}", 'w+') as file:
        file.write("\\n".join([str(var) for var in web_var.standard_vars.items()]))
"""


class TestRerunJobsPageIntegrationMultiVar(BaseTestCase, NavbarTestMixin, FooterTestMixin, AccessibilityTestMixin):
    fixtures = BaseTestCase.fixtures + ["run_with_multiple_variables"]

    accessibility_test_ignore_rules = {
        # https://github.com/ISISScientificComputing/autoreduce/issues/1267
        "duplicate-id-aria": "input",
    }

    @classmethod
    def setUpClass(cls):
        """Starts external services and sets instrument for all test cases"""
        super().setUpClass()
        cls.instrument_name = "TestInstrument"
        cls.data_archive, cls.queue_client, cls.listener = setup_external_services(cls.instrument_name, 21, 21)
        cls.data_archive.add_reduction_script(cls.instrument_name, SCRIPT)
        cls.data_archive.add_reduce_vars_script(
            cls.instrument_name, """standard_vars={"variable_str":"test_variable_value_123",
                                                "variable_int":123, "variable_float":123.321,
                                                "variable_listint":[1,2,3], "variable_liststr":["a","b","c"],
                                                "variable_none":None, "variable_empty":"", "variable_bool":True}""")
        cls.instrument_name = "TestInstrument"
        cls.rb_number = 1234567
        cls.run_number = 99999

    @classmethod
    def tearDownClass(cls) -> None:
        """Stops external services."""
        cls.queue_client.disconnect()
        cls.data_archive.delete()
        super().tearDownClass()

    def setUp(self) -> None:
        """Sets up RerunJobsPage before each test case"""
        super().setUp()
        self.page = RerunJobsPage(self.driver, self.instrument_name)
        self.page.launch()

    def test_variables_appear_as_expected(self):
        """
        Test: Just opening the submit page and clicking rerun
        """
        assert self.page.variable_str_field_val == "value1"
        assert self.page.variable_int_field_val == "123"
        assert self.page.variable_float_field_val == "123.321"
        assert self.page.variable_listint_field_val == "[1, 2, 3]"
        assert self.page.variable_liststr_field_val == "['a', 'b', 'c']"
        assert self.page.variable_none_field_val == "None"
        assert self.page.variable_bool_field_val == "True"

    def test_submit_rerun_same_variables(self):
        """
        Test: Just opening the submit page and clicking rerun
        """
        result = submit_and_wait_for_result(self)
        assert len(result) == 2

        assert result[0].run_version == 0
        assert result[1].run_version == 1

        for run0_var, run1_var in zip(result[0].run_variables.all(), result[1].run_variables.all()):
            assert run0_var.variable == run1_var.variable

    def test_submit_rerun_changed_variable_arbitrary_value(self):
        """
        Test: Open submit page, change a variable, submit the run
        """
        new_str_value = "the new value in the field"
        self.page.variable_str_field = new_str_value
        new_int = "42"
        self.page.variable_int_field = new_int
        new_float = "144.33"
        self.page.variable_float_field = new_float
        new_listint = "[111, 222]"
        self.page.variable_listint_field = new_listint
        new_liststr = "['string1', 'string2']"
        self.page.variable_liststr_field = new_liststr
        new_bool = False
        self.page.variable_bool_field = new_bool

        result = submit_and_wait_for_result(self)
        assert len(result) == 2

        assert result[0].run_version == 0
        assert result[1].run_version == 1

        for run0_var, run1_var in zip(result[0].run_variables.all(), result[1].run_variables.all()):
            # the value of the variable has been overwritten because it's the same run number
            assert run0_var.variable == run1_var.variable

        run_vars = result[1].run_variables.all()
        assert run_vars[0].variable.value == new_str_value
        assert run_vars[1].variable.value == new_int
        assert run_vars[2].variable.value == new_float
        assert run_vars[3].variable.value == new_listint
        assert run_vars[4].variable.value == new_liststr
        assert run_vars[5].variable.value is None
        assert run_vars[6].variable.value == ""
        assert run_vars[7].variable.value == "False"

        with open(TEMP_OUT_FILE.name, 'r') as fil:
            contents = fil.read()

        for runvar in run_vars:
            assert runvar.variable.name in contents
            assert str(runvar.variable.value) in contents



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
WSGI config for autoreduce_django project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/dev/howto/deployment/wsgi/
"""

import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "autoreduce_webapp.autoreduce_django.settings")
# pylint: disable=invalid-name
application = get_wsgi_application()



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Selenium tests for the overview page
"""

from autoreduce_webapp.selenium_tests.pages.overview_page import OverviewPage
from autoreduce_webapp.selenium_tests.tests.base_tests import (NavbarTestMixin, BaseTestCase, FooterTestMixin,
                                                               AccessibilityTestMixin)


class TestOverviewPage(NavbarTestMixin, BaseTestCase, FooterTestMixin, AccessibilityTestMixin):
    """
    Test cases for the overview page
    """
    fixtures = BaseTestCase.fixtures + ["test_overview_page"]

    def setUp(self) -> None:
        """
        Sets up the OverviewPage object
        """
        super().setUp()
        self.page = OverviewPage(self.driver)

    def test_correct_instruments_visible(self):
        """
        Tests: Correct instruments displayed
        """
        actual_instruments = self.page.launch().get_instruments_from_buttons()
        expected_instruments = {"ActiveInstrument", "InactiveInstrument", "PausedInstrument"}
        self.assertTrue(expected_instruments.issubset(actual_instruments))

    def test_instrument_buttons_go_to_instrument_summary_pages(self):
        """
        Tests: Instrument overviewpage is navigated to
        When: Instrument button is clicked
        """
        instruments = self.page.launch().get_instruments_from_buttons()
        for instrument in instruments:
            instrument_page = self.page.click_instrument(instrument)
            self.assertTrue(self.driver.current_url.endswith(instrument_page.url_path()))
            self.page.launch()

    def test_tour(self):
        """
        Tests: Tour run through on overview page
        """
        self.page.launch().start_tour()
        self.assertTrue(self.page.is_tour_visible())
        self.assertFalse(self.page.is_tour_previous_button_enabled())
        self.page.next_tour_step()
        self.assertTrue(self.page.is_tour_visible())
        self.page.next_tour_step()
        self.assertTrue(self.page.is_tour_visible())
        self.assertFalse(self.page.is_tour_next_button_enabled())
        self.page.end_tour()
        self.assertTrue(self.page.is_tour_hidden())



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Taken from: https://djangosnippets.org/snippets/1568/
"""
from django.conf import settings
from django.template import Library, Node, TemplateSyntaxError, Variable
from django.urls import get_callable, URLResolver

# pylint:disable=invalid-name
register = Library()


class ViewNode(Node):
    """
    Handles the ViewNode
    """
    def __init__(self, url_or_view, args, kwargs):
        self.url_or_view = url_or_view
        self.args = args
        self.kwargs = kwargs

    def render(self, context):
        """
        Render the view node
        """
        if 'request' not in context:
            return ""
        request = context['request']

        url_or_view = Variable(self.url_or_view).resolve(context)
        try:
            urlconf = getattr(request, "urlconf", settings.ROOT_URLCONF)
            resolver = URLResolver(r'^/', urlconf)
            view, args, kwargs = resolver.resolve(url_or_view)
        # pylint: disable=bare-except
        except:
            view = get_callable(url_or_view)
            args = [Variable(arg).resolve(context) for arg in self.args]
            kwargs = {}
            for key, value in self.kwargs.items():
                kwargs[key] = Variable(value).resolve(context)

        try:
            if callable(view):
                render_fnc = view(context['request'], *args, **kwargs).content.\
                    decode('unicode_escape')
                return render_fnc
            raise "%r is not callable" % view
        # pylint: disable=bare-except
        except:
            if settings.DEBUG_PROPAGATE_EXCEPTIONS:
                raise
        return ""


def do_view(_, token):
    """
    Inserts the output of a view, using fully qualified view name (and then some
    args), a or local Django URL.

     {% view view_or_url arg[ arg2] k=v [k2=v2...] %}

    This might be helpful if you are trying to do 'on-server' AJAX of page
    panels. Most browsers can call back to the server to get panels of content
    asynchonously, whilst others (such as mobiles that don't support AJAX very
    well) can have a template that embeds the output of the URL synchronously
    into the main page. Yay! Go the mobile web!

    Follow standard templatetag instructions for installing.

    IMPORTANT: the calling template must receive a context variable called
    'request' containing the original HttpRequest. This means you should be OK
    with permissions and other session state.

    ALSO NOTE: that middleware is not invoked on this 'inner' view.

    Example usage...

    Using a view name (or something that evaluates to a view name):
     {% view "mymodule.views.inner" "value" %}
     {% view "mymodule.views.inner" keyword="value" %}
     {% view "mymodule.views.inner" arg_expr %}
     {% view "mymodule.views.inner" keyword=arg_expr %}
     {% view view_expr "value" %}
     {% view view_expr keyword="value" %}
     {% view view_expr arg_expr %}
     {% view view_expr keyword=arg_expr %}

    Using a URL (or something that evaluates to a URL):
     {% view "/inner" %}
     {% view url_expr %}


    (Note that every argument will be evaluated against context except for the
    names of any keyword arguments. If you're warped enough to need evaluated
    keyword names, then you're probably smart enough to add this yourself!)

    """

    args = []
    kwargs = {}
    tokens = token.split_contents()
    if len(tokens) < 2:
        raise TemplateSyntaxError("%r tag requires one or more arguments" % token.contents.split()[0])
    _ = tokens.pop(0)
    url_or_view = tokens.pop(0)
    # pylint: disable=redefined-argument-from-local
    for token in tokens:
        equals = token.find("=")
        if equals == -1:
            args.append(token)
        else:
            kwargs[str(token[:equals])] = token[equals + 1:]
    return ViewNode(url_or_view, args, kwargs)


register.tag('view', do_view)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #

from django.urls import reverse
from selenium.webdriver.remote.webelement import WebElement

from autoreduce_webapp.selenium_tests.pages.component_mixins.footer_mixin import FooterMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.navbar_mixin import NavbarMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.tour_mixin import TourMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.rerun_form_mixin import RerunFormMixin
from autoreduce_webapp.selenium_tests.pages.page import Page


class ConfigureNewRunsPage(Page, RerunFormMixin, NavbarMixin, FooterMixin, TourMixin):
    """
    Page model class for run summary page
    """
    def __init__(self, driver, instrument, run_start=None, experiment_reference=None):
        super().__init__(driver)
        self.instrument = instrument
        self._run_start_number = run_start
        self._experiment_reference = experiment_reference

    def url_path(self) -> str:
        """
        Return the current URL of the page.
        :return: (str) the url path
        """
        kwargs = {
            "instrument": self.instrument,
        }
        if self._experiment_reference:
            kwargs["experiment_reference"] = self._experiment_reference
            return reverse("instrument:variables_by_experiment", kwargs=kwargs)
        else:
            if self._run_start_number:
                kwargs["start"] = self._run_start_number
            return reverse("instrument:variables", kwargs=kwargs)

    @property
    def run_start(self) -> WebElement:
        """
        Return the run start input WebElement
        """
        return self.driver.find_element_by_id("run_start")

    @property
    def run_start_val(self) -> WebElement:
        """Return the value of the run start WebElement"""
        return self.driver.find_element_by_id("run_start").get_attribute("value")

    @run_start.setter
    def run_start(self, value):
        """Set the value of the start run"""
        self._set_field(self.run_start, value)

    @property
    def run_end(self) -> WebElement:
        """Return the end run WebElement"""
        return self.driver.find_element_by_id("run_end")

    @property
    def run_end_val(self) -> WebElement:
        """Return the value of the run end element"""
        return self.driver.find_element_by_id("run_end").get_attribute("value")

    @run_end.setter
    def run_end(self, value):
        """Set the value of the end run"""
        self._set_field(self.run_end, value)

    @property
    def experiment_reference_number(self) -> WebElement:
        """Return the reference number input WebElement"""
        return self.driver.find_element_by_id("experiment_reference_number")

    @property
    def experiment_reference_number_val(self) -> WebElement:
        """Return the value of the reference number input WebElement"""
        return self.driver.find_element_by_id("experiment_reference_number").get_attribute("value")

    @experiment_reference_number.setter
    def experiment_reference_number(self, value):
        """Set the reference number value for the reference number input"""
        self._set_field(self.experiment_reference_number, value)

    @property
    def range_or_experiment_toggle(self) -> WebElement:
        """Return the run range or experiment reference toggle"""
        return self.driver.find_element_by_id("variable-range-toggle")

    @property
    def reset_to_current_values(self) -> WebElement:
        """
        Finds and returns the "Reset to values in the current reduce_vars script" button
        """
        return self.driver.find_element_by_id("currentScript")

    @property
    def replace_confirm(self) -> WebElement:
        """Return the confirmation button from the modal"""
        return self.driver.find_element_by_id("replace_confirm")

    @property
    def go_to_other(self) -> WebElement:
        """Return the link to toggle between by run range or by reference"""
        return self.driver.find_element_by_id("go_to_other")



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
View functions for displaying Variable data
This imports into another view, thus no middleware
"""
import logging

from django.shortcuts import redirect, render
from autoreduce_db.instrument.models import InstrumentVariable
from autoreduce_db.reduction_viewer.models import Instrument, ReductionRun
from autoreduce_qp.queue_processor.variable_utils import VariableUtils

from autoreduce_webapp.autoreduce_django.view_utils import (check_permissions, login_and_uows_valid, render_with)
from autoreduce_webapp.reduction_viewer.utils import ReductionRunUtils

LOGGER = logging.getLogger("app")


# pylint:disable=too-many-locals
def summarize_variables(request, instrument, last_run_object):
    """
    Handles view request for the instrument summary page
    """
    # pylint:disable=no-member
    instrument = Instrument.objects.get(name=instrument)

    # pylint:disable=invalid-name
    current_variables = [runvar.variable.instrumentvariable for runvar in last_run_object.run_variables.all()]

    upcoming_variables_by_run = InstrumentVariable.objects.filter(start_run__gt=last_run_object.run_number,
                                                                  instrument=instrument)
    upcoming_variables_by_experiment = InstrumentVariable.objects.filter(
        experiment_reference__gte=last_run_object.experiment.reference_number, instrument=instrument)

    # There's a known issue with inaccurate display of tracks script:
    # https://github.com/ISISScientificComputing/autoreduce/issues/1187
    # Creates a nested dictionary for by-run
    upcoming_variables_by_run_dict = {}
    for variable in upcoming_variables_by_run:
        if variable.start_run not in upcoming_variables_by_run_dict:
            upcoming_variables_by_run_dict[variable.start_run] = {
                'run_start': variable.start_run,
                'run_end': 0,  # We'll fill this in after
                'tracks_script': variable.tracks_script,
                'variables': [],
                'instrument': instrument,
            }
        upcoming_variables_by_run_dict[variable.start_run]['variables'].append(variable)

    # Fill in the run end numbers
    run_end = 0
    for run_number in sorted(upcoming_variables_by_run_dict.keys(), reverse=True):
        upcoming_variables_by_run_dict[run_number]['run_end'] = run_end
        run_end = max(run_number - 1, 0)

    if current_variables:
        current_start = current_variables[0].start_run
        next_run_starts = list(
            filter(lambda start: start > current_start, sorted(upcoming_variables_by_run_dict.keys())))
        current_end = next_run_starts[0] - 1 if next_run_starts else 0

        current_vars = {
            'run_start': current_start,
            'run_end': current_end,
            'tracks_script': not any((var.tracks_script for var in current_variables)),
            'variables': current_variables,
            'instrument': instrument,
        }
    else:
        current_vars = {}

    # Move the upcoming vars into an ordered list
    upcoming_variables_by_run_ordered = []
    for key in sorted(upcoming_variables_by_run_dict):
        upcoming_variables_by_run_ordered.append(upcoming_variables_by_run_dict[key])

    # Create a nested dictionary for by-experiment
    upcoming_variables_by_experiment_dict = {}
    for variables in upcoming_variables_by_experiment:
        if variables.experiment_reference not in upcoming_variables_by_experiment_dict:
            upcoming_variables_by_experiment_dict[variables.experiment_reference] = {
                'experiment': variables.experiment_reference,
                'variables': [],
                'instrument': instrument,
            }
        upcoming_variables_by_experiment_dict[variables.experiment_reference]['variables'].\
            append(variables)

    # Move the upcoming vars into an ordered list
    upcoming_variables_by_experiment_ordered = []
    for key in sorted(upcoming_variables_by_experiment_dict):
        upcoming_variables_by_experiment_ordered.append(upcoming_variables_by_experiment_dict[key])
    sorted(upcoming_variables_by_experiment_ordered, key=lambda r: r['experiment'])

    context_dictionary = {
        'instrument': instrument,
        'current_variables': current_vars,
        'upcoming_variables_by_run': upcoming_variables_by_run_ordered,
        'upcoming_variables_by_experiment': upcoming_variables_by_experiment_ordered,
    }

    return render(request, 'snippets/instrument_summary_variables.html', context_dictionary)


# pylint:disable=unused-argument
@login_and_uows_valid
@check_permissions
def delete_instrument_variables(request, instrument=None, start=0, end=0, experiment_reference=None):
    """
    Handles request for deleting instrument variables

    :param instrument: Name of the instrument for which variables are being deleted
    :param start: The start run from which variables are being deleted
    :param end: Used to limit how many variables get deleted, otherwise a delete would wipe ALL variables > start
    :param experiment_reference: If provided - use the experiment reference to delete variables instead of start_run
    """

    # We "save" an empty list to delete the previous variables.
    if experiment_reference is not None:
        InstrumentVariable.objects.filter(instrument__name=instrument,
                                          experiment_reference=experiment_reference).delete()
    else:
        start_run_kwargs = {"start_run__gte": start}
        if end > 0:
            start_run_kwargs["start_run__lte"] = end
        InstrumentVariable.objects.filter(instrument__name=instrument, **start_run_kwargs).delete()

    return redirect('instrument:variables_summary', instrument=instrument)


@login_and_uows_valid
@check_permissions
@render_with('variables_summary.html')
# pylint:disable=no-member
def instrument_variables_summary(request, instrument):
    """
    Handles request to view instrument variables
    """
    instrument = Instrument.objects.get(name=instrument)
    context_dictionary = {'instrument': instrument, 'last_instrument_run': instrument.reduction_runs.last()}
    return context_dictionary


@login_and_uows_valid
@check_permissions
@render_with('snippets/edit_variables.html')
def current_default_variables(request, instrument=None):
    """
    Handles request to view default variables
    """

    try:
        current_variables = VariableUtils.get_default_variables(instrument)
    except (FileNotFoundError, ImportError, SyntaxError) as err:
        return {"message": str(err)}
    standard_vars = current_variables["standard_vars"]
    advanced_vars = current_variables["advanced_vars"]

    context_dictionary = {
        'instrument': instrument,
        'standard_variables': standard_vars,
        'advanced_variables': advanced_vars,
    }
    return context_dictionary


def render_run_variables(request, instrument_name, run_number, run_version=0):
    """
    Handles request to view the summary of a run
    """
    # pylint:disable=no-member
    reduction_run = ReductionRun.objects.get(instrument__name=instrument_name,
                                             run_number=run_number,
                                             run_version=run_version)

    vars_kwargs = ReductionRunUtils.make_kwargs_from_runvariables(reduction_run)
    standard_vars = vars_kwargs["standard_vars"]
    advanced_vars = vars_kwargs["advanced_vars"]

    try:
        current_variables = VariableUtils.get_default_variables(instrument_name)
        current_standard_variables = current_variables["standard_vars"]
        current_advanced_variables = current_variables["advanced_vars"]
    except (FileNotFoundError, ImportError, SyntaxError):
        current_standard_variables = {}
        current_advanced_variables = {}

    context_dictionary = {
        'run_number': run_number,
        'run_version': run_version,
        'standard_variables': standard_vars,
        'advanced_variables': advanced_vars,
        'current_standard_variables': current_standard_variables,
        'current_advanced_variables': current_advanced_variables,
        'instrument': reduction_run.instrument,
    }
    return render(request, 'snippets/run_variables.html', context_dictionary)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Asserts the variable type of tables
"""
from django.template import Library

# pylint:disable=invalid-name
register = Library()


@register.simple_tag
def variable_type(var_type):
    """
    :return: variable type expected deepening on table element
    """
    if var_type == 'boolean':
        return 'checkbox'
    if var_type in ('list_number', 'list_text'):
        return 'text'
    return var_type



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module for the accessibility statement page model
"""
from django.urls.base import reverse
from selenium.webdriver.remote.webelement import WebElement

from autoreduce_webapp.selenium_tests.pages.component_mixins.footer_mixin import FooterMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.navbar_mixin import NavbarMixin
from autoreduce_webapp.selenium_tests.pages.page import Page


class AccessibilityStatementPage(Page, NavbarMixin, FooterMixin):
    """
    Page model class for accessibility statement page
    """
    @staticmethod
    def url_path() -> str:
        """
        Return the current URL of the page.
        :return: (str) the url path
        """
        return reverse("accessibility_statement")

    def _get_accessibility_statement_contents_element(self) -> WebElement:
        """
        Get the <div> #accessibility-statement-contents
        :return: (WebElement) The element #accessibility-statement-contents
        """
        return self.driver.find_element_by_id("accessibility-statement-contents")

    def is_accessibility_statement_visible(self) -> bool:
        """
        Is #accessibility-statement-contents visible
        :return: (str) True if #accessibility-statement-contents is visible else False
        """
        return self._get_accessibility_statement_contents_element().is_displayed()

    def get_accessibility_statement_contents_text(self) -> str:
        """
        Get the contents of #accessibility-statement-contents
        :return: (str) The text in #accessibility-statement-contents
        """
        return self._get_accessibility_statement_contents_element().text



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module for the help summary page model
"""
from typing import List

from django.urls.base import reverse
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.remote.webelement import WebElement

from autoreduce_webapp.selenium_tests.pages.component_mixins.footer_mixin import FooterMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.navbar_mixin import NavbarMixin
from autoreduce_webapp.selenium_tests.pages.page import Page


class HelpPage(Page, NavbarMixin, FooterMixin):
    """
    Page model class for help page
    """
    @staticmethod
    def url_path() -> str:
        """
        Return the current URL of the page.
        :return: (str) the url path
        """
        return reverse("help")

    def get_sidenav_contents_elements(self) -> List[WebElement]:
        """
        Get the contents of the sidenav.
        :return: (List) A list of <li> WebElements in #sidenav-contents
        """
        return self.driver.find_elements_by_xpath("//ul[@id='sidenav-contents']/li")

    def get_sidenav_contents(self) -> List[str]:
        """
        Get the contents of the sidenav as a list of strings.
        :return: (List) A list of strings inside of each <a> element
        """
        return [x.find_element_by_tag_name("a").text.strip() for x in self.get_sidenav_contents_elements()]

    def get_help_topic_elements(self) -> List[WebElement]:
        """
        Get the help topics
        :return: (List) A list of <section> WebElements of class .help-topic
        """
        return self.driver.find_elements_by_class_name("help-topic")

    def get_each_help_topic_category(self) -> List[str]:
        """
        Get each data-category value from each help topic
        :return: (List) A list of categories from each help topic
        """
        return [x.get_attribute("data-category") for x in self.get_help_topic_elements()]

    def get_each_help_topic_content(self) -> List[str]:
        """
        Get each content text for each help topic
        :return: (List) A list of topic contents
        """
        return [x.find_element_by_class_name("panel-body").text for x in self.get_help_topic_elements()]

    def get_help_topic_header_elements(self) -> List[WebElement]:
        """
        Get the help topic headers
        :return: (List) A list of <div> WebElements of class .panel-header
        """
        return [x.find_element_by_xpath("./div[@class='panel-heading']") for x in self.get_help_topic_elements()]

    def get_help_topic_header_link_elements(self) -> List[WebElement]:
        """
        Get the help topic headers link elements
        :return: (List) A list of WebElement links found in the headers of help topics
        """
        return [x.find_element_by_xpath("./h3/a") for x in self.get_help_topic_header_elements()]

    def get_each_help_topic_header(self) -> List[str]:
        """
        Get the headers of the help topics.
        :return (List) A list of help topic headers
        """
        return [x.find_element_by_xpath("./h3/a").text.strip() for x in self.get_help_topic_header_elements()]

    def get_category_filter_elements(self) -> List[WebElement]:
        """
        Get the category filter elements in #category-filter
        :return: (List) A list of filter elements
        """
        return self.driver.find_elements_by_xpath("//div[@id='category-filter']/label")

    def get_help_topic_filters(self) -> List[str]:
        """
        Get the help topic filters from the category filter
        :return: (List) A list of categories
        """
        return [x.get_attribute("data-category").strip() for x in self.get_category_filter_elements()]

    def get_available_help_topic_categories(self) -> List[str]:
        """
        Get the available categories for topics on the page
        :return: (List) A list of categories
        """
        filters = self.get_help_topic_filters()
        filters.remove("all")
        return filters

    def click_category_filter(self, category):
        """
        Click a category filter button
        :param category: (str) The category name to select
        """
        filter_button = next(
            (x for x in self.get_category_filter_elements() if x.get_attribute("data-category") == category), None)

        if filter_button is None:
            raise Exception(f"Category {category} does not exist.")

        filter_button.click()

    def filter_help_topics_by_search_term(self, text):
        """
        Filter help topics by a search term using fuzzy search
        :param text: The search term to enter into #help-search
        """
        help_search = self.driver.find_element_by_id("help-search")
        help_search.clear()
        help_search.send_keys(text)

    def clear_search_box(self):
        """
        Clear the text in the search box
        """
        help_search = self.driver.find_element_by_id("help-search")
        help_search.clear()
        help_search.send_keys(Keys.BACKSPACE)  # Send onKeyUp event to #help-search

    def get_visible_help_topic_elements(self) -> List[WebElement]:
        """
        Get the visible help topics on the page
        :return: (List) A list of help topics that are not display:none
        """
        return [x for x in self.get_help_topic_elements() if x.is_displayed()]



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
The view for the django database model

This file contains a large amount of pylint disables as there are no
working unit tests for this code at the point of correcting the pylint errors
as such we should remove the pylint disables and fix the code when we
can be more confident we are not affecting the execution
"""
# pylint:disable=imported-auth-user
import json
import logging
import operator
import traceback

from autoreduce_db.reduction_viewer.models import (Experiment, Instrument, ReductionRun, Status)
from autoreduce_qp.queue_processor.variable_utils import VariableUtils
from django.contrib.auth import authenticate, login
from django.contrib.auth import logout as django_logout
from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.db.models import Q
from django.http import HttpResponseNotFound
from django.shortcuts import redirect
from django.utils.http import url_has_allowed_host_and_scheme

from autoreduce_webapp.autoreduce_django.icat_cache import (ICATCache, ICATConnectionException)
from autoreduce_webapp.autoreduce_django.settings import (ALLOWED_HOSTS, DEVELOPMENT_MODE, UOWS_LOGIN_URL,
                                                          USER_ACCESS_CHECKS)
from autoreduce_webapp.autoreduce_django.uows_client import UOWSClient
from autoreduce_webapp.autoreduce_django.view_utils import (check_permissions, login_and_uows_valid, render_with,
                                                            require_admin)
from autoreduce_webapp.autoreduce_django.views import render_error
from autoreduce_webapp.plotting.plot_handler import PlotHandler
from autoreduce_webapp.reduction_viewer.utils import ReductionRunUtils
from autoreduce_webapp.reduction_viewer.view_utils import (deactivate_invalid_instruments, get_interactive_plot_data,
                                                           make_data_analysis_url)
from autoreduce_webapp.utilities.pagination import CustomPaginator

LOGGER = logging.getLogger('app')


@deactivate_invalid_instruments
def index(request):
    """
    Render the index page
    """
    return_url = _make_return_url(request, request.GET.get('next'))

    use_query_next = request.build_absolute_uri(request.GET.get('next'))
    default_next = 'overview'

    authenticated = False

    if DEVELOPMENT_MODE:
        user = authenticate(username="super", password="super", backend="django.contrib.auth.backends.ModelBackend")
        login(request, user)
        authenticated = True
    else:
        if 'sessionid' in request.session.keys():
            authenticated = request.user.is_authenticated \
                            and UOWSClient().check_session(request.session['sessionid'])

    if authenticated:
        if request.GET.get('next'):
            return_url = use_query_next
        else:
            return_url = default_next
    elif request.GET.get('sessionid'):
        request.session['sessionid'] = request.GET.get('sessionid')
        try:
            user = authenticate(token=request.GET.get('sessionid'))
        except ICATConnectionException as excep:
            return render_error(request, str(excep))

        if user is not None:
            if user.is_active:
                login(request, user)
                if request.GET.get('next'):
                    return_url = use_query_next
                else:
                    return_url = default_next

    return redirect(return_url)


def _make_return_url(request, next_url):
    """
    Make the return URL based on whether a next_url is present in the url.

    If there is a next_url, verify that the url is safe and allowed before using it. If not, default to the host.
    """
    if next_url:
        if url_has_allowed_host_and_scheme(next_url, ALLOWED_HOSTS, require_https=True):
            return UOWS_LOGIN_URL + request.build_absolute_uri(next_url)
        else:
            # the next_url was not safe so don't use it - build from request.path to ignore GET parameters
            return UOWS_LOGIN_URL + request.build_absolute_uri(request.path)
    else:
        return UOWS_LOGIN_URL + request.build_absolute_uri()


@login_and_uows_valid
def logout(request):
    """
    Render the logout page
    """
    session_id = request.session.get('sessionid')
    if session_id:
        UOWSClient().logout(session_id)
    django_logout(request)
    request.session.flush()
    return redirect('overview')


@login_and_uows_valid
@render_with('overview.html')
# pylint:disable=no-member
def overview(_):
    """
    Render the overview landing page (redirect from /index)
    Note: _ is replacing the passed in request parameter
    """
    context_dictionary = {}
    instruments = Instrument.objects.values_list("name", flat=True)
    if instruments:
        context_dictionary = {'instrument_list': instruments}
    return context_dictionary


@login_and_uows_valid
@render_with('run_queue.html')
# pylint:disable=no-member
def run_queue(request):
    """
    Render status of queue
    """
    # Get all runs that should be shown
    queued_status = Status.get_queued()
    processing_status = Status.get_processing()
    pending_jobs = ReductionRun.objects.filter(Q(status=queued_status)
                                               | Q(status=processing_status)).order_by('created')
    # Filter those which the user shouldn't be able to see
    if USER_ACCESS_CHECKS and not request.user.is_superuser:
        try:
            with ICATCache(AUTH='uows', SESSION={'sessionid': request.session['sessionid']}) as icat:
                pending_jobs = filter(lambda job: job.experiment.reference_number in icat.get_associated_experiments(
                    int(request.user.username)), pending_jobs)  # check RB numbers
                pending_jobs = filter(
                    lambda job: job.instrument.name in icat.get_owned_instruments(int(request.user.username)),
                    pending_jobs)  # check instrument
        except ICATConnectionException as excep:
            return render_error(request, str(excep))
    # Initialise list to contain the names of user/team that started runs
    started_by = []
    # cycle through all filtered runs and retrieve the name of the user/team that started the run
    for run in pending_jobs:
        started_by.append(started_by_id_to_name(run.started_by))
    # zip the run information with the user/team name to enable simultaneous iteration with django
    context_dictionary = {'queue': zip(pending_jobs, started_by)}
    return context_dictionary


@require_admin
@login_and_uows_valid
@render_with('fail_queue.html')
# pylint:disable=no-member,too-many-locals
def fail_queue(request):
    """
    Render status of failed queue
    """
    # render the page
    error_status = Status.get_error()
    failed_jobs = ReductionRun.objects.filter(Q(status=error_status)
                                              & Q(hidden_in_failviewer=False)).order_by('-created')
    context_dictionary = {
        'queue': failed_jobs,
        'status_success': Status.get_completed(),
        'status_failed': Status.get_error()
    }

    if request.method == 'POST':
        # perform the specified action
        action = request.POST.get("action", "default")
        selected_run_string = request.POST.get("selectedRuns", [])
        selected_runs = json.loads(selected_run_string)
        try:
            for run in selected_runs:
                run_number = int(run[0])
                run_version = int(run[1])

                reduction_run = failed_jobs.get(run_number=run_number, run_version=run_version)

                if action == "hide":
                    reduction_run.hidden_in_failviewer = True
                    reduction_run.save()

                elif action == "rerun":
                    highest_version = max([int(runL[1]) for runL in selected_runs if int(runL[0]) == run_number])
                    if run_version != highest_version:
                        continue  # do not run multiples of the same run

                    ReductionRunUtils.send_retry_message_same_args(request.user.id, reduction_run)

                elif action == "default":
                    pass

        # pylint:disable=broad-except
        except Exception as exception:
            fail_str = "Selected action failed: %s %s" % (type(exception).__name__, exception)
            LOGGER.info("Failed to carry out fail_queue action - %s", fail_str)
            context_dictionary["message"] = fail_str

    return context_dictionary


@login_and_uows_valid
@check_permissions
@render_with('run_summary.html')
# pylint:disable=no-member,too-many-locals
def run_summary(_, instrument_name=None, run_number=None, run_version=0):
    """
    Render run summary
    """
    # pylint:disable=broad-except
    try:
        history = ReductionRun.objects.filter(instrument__name=instrument_name, run_number=run_number).order_by(
            '-run_version').select_related('status').select_related('experiment').select_related('instrument')
        if len(history) == 0:
            raise ValueError(f"Could not find any matching runs for instrument {instrument_name} run {run_number}")

        run = next(run for run in history if run.run_version == int(run_version))
        started_by = started_by_id_to_name(run.started_by)
        # run status value of "s" means the run is skipped
        is_skipped = run.status.value == "s"
        is_rerun = len(history) > 1

        location_list = run.reduction_location.all()
        reduction_location = ""
        if location_list:
            reduction_location = location_list[0].file_path
        if reduction_location and '\\' in reduction_location:
            reduction_location = reduction_location.replace('\\', '/')

        data_analysis_link_url = make_data_analysis_url(reduction_location) if reduction_location else ""

        rb_number = run.experiment.reference_number
        try:
            current_variables = VariableUtils.get_default_variables(run.instrument.name)
        except (FileNotFoundError, ImportError, SyntaxError):
            current_variables = {}

        has_reduce_vars = bool(current_variables)
        has_run_variables = bool(run.run_variables.count())
        context_dictionary = {
            'run': run,
            'run_number': run_number,
            'instrument': instrument_name,
            'run_version': run_version,
            'is_skipped': is_skipped,
            'is_rerun': is_rerun,
            'history': history,
            'reduction_location': reduction_location,
            'started_by': started_by,
            'has_reduce_vars': has_reduce_vars,
            'has_run_variables': has_run_variables,
            'data_analysis_link_url': data_analysis_link_url
        }

    except PermissionDenied:
        raise
    except Exception as exception:
        # Error that we cannot recover from - something wrong with instrument, run, or experiment
        LOGGER.error(exception)
        return {"message": str(exception)}

    if reduction_location:
        try:
            plot_handler = PlotHandler(data_filepath=run.data_location.first().file_path,
                                       server_dir=reduction_location,
                                       rb_number=rb_number)
            local_plot_locs, server_plot_locs = plot_handler.get_plot_file()
            if local_plot_locs:
                context_dictionary['static_plots'] = [
                    location for location in local_plot_locs if not location.endswith(".json")
                ]

                context_dictionary['interactive_plots'] = get_interactive_plot_data(server_plot_locs)
        except Exception as exception:
            # Lack of plot images is recoverable - we shouldn't stop the whole page rendering
            # if something is wrong with the plot images - but display an error message
            err_msg = "Encountered error while retrieving plots for this run"
            LOGGER.error("%s. Instrument: %s, run %s. RB Number %s Error: %s", err_msg, run.instrument.name, run,
                         rb_number, exception)
            context_dictionary["plot_error_message"] = f"{err_msg}."

    return context_dictionary


@login_and_uows_valid
@check_permissions
@render_with('runs_list.html')
# pylint:disable=no-member,unused-argument,too-many-locals
def runs_list(request, instrument=None):
    """
    Render instrument summary
    """
    try:
        filter_by = request.GET.get('filter', 'run')
        instrument_obj = Instrument.objects.get(name=instrument)
    except Instrument.DoesNotExist:
        return {'message': "Instrument not found."}

    try:
        sort_by = request.GET.get('sort', 'run')
        if sort_by == 'run':
            runs = (ReductionRun.objects.only('status', 'last_updated', 'run_number', 'run_version',
                                              'run_description').select_related('status').filter(
                                                  instrument=instrument_obj).order_by('-run_number', 'run_version'))
        else:
            runs = (ReductionRun.objects.only(
                'status', 'last_updated', 'run_number', 'run_version',
                'run_description').select_related('status').filter(instrument=instrument_obj).order_by('-last_updated'))

        if len(runs) == 0:
            return {'message': "No runs found for instrument."}

        try:
            current_variables = VariableUtils.get_default_variables(instrument_obj.name)
            error_reason = ""
        except FileNotFoundError:
            current_variables = {}
            error_reason = "reduce_vars.py is missing for this instrument"
        except (ImportError, SyntaxError):
            current_variables = {}
            error_reason = "reduce_vars.py has an import or syntax error"

        has_variables = bool(current_variables)

        context_dictionary = {
            'instrument': instrument_obj,
            'instrument_name': instrument_obj.name,
            'runs': runs,
            'last_instrument_run': runs[0],
            'processing': runs.filter(status=Status.get_processing()),
            'queued': runs.filter(status=Status.get_queued()),
            'filtering': filter_by,
            'sort': sort_by,
            'has_variables': has_variables,
            'error_reason': error_reason
        }

        if filter_by == 'experiment':
            experiments_and_runs = {}
            experiments = Experiment.objects.filter(reduction_runs__instrument=instrument_obj). \
                order_by('-reference_number').distinct()
            for experiment in experiments:
                associated_runs = runs.filter(experiment=experiment). \
                    order_by('-created')
                experiments_and_runs[experiment] = associated_runs
            context_dictionary['experiments'] = experiments_and_runs
        else:
            max_items_per_page = request.GET.get('pagination', 10)
            custom_paginator = CustomPaginator(page_type=sort_by,
                                               query_set=runs,
                                               items_per_page=max_items_per_page,
                                               page_tolerance=3,
                                               current_page=request.GET.get('page', 1))
            context_dictionary['paginator'] = custom_paginator
            context_dictionary['last_page_index'] = len(custom_paginator.page_list)
            context_dictionary['max_items'] = max_items_per_page

    # pylint:disable=broad-except
    except Exception:
        LOGGER.error(traceback.format_exc())
        return {'message': "An unexpected error has occurred when loading the instrument."}

    return context_dictionary


@login_and_uows_valid
@check_permissions
@render_with('experiment_summary.html')
# pylint:disable=no-member,too-many-locals
def experiment_summary(request, reference_number=None):
    """
    Render experiment summary
    """
    try:
        experiment = Experiment.objects.get(reference_number=reference_number)
        runs = ReductionRun.objects.filter(experiment=experiment).order_by('-run_version')
        data = []
        reduced_data = []
        started_by = []
        for run in runs:
            for location in run.data_location.all():
                if location not in data:
                    data.append(location)
            for location in run.reduction_location.all():
                if location not in reduced_data:
                    reduced_data.append(location)
            started_by.append(started_by_id_to_name(run.started_by))
        sorted_runs = sorted(runs, key=operator.attrgetter('last_updated'), reverse=True)
        runs_with_started_by = zip(sorted_runs, started_by)

        try:
            if DEVELOPMENT_MODE:
                # If we are in development mode use user/password for ICAT from django settings
                # e.g. do not attempt to use same authentication as the user office
                try:
                    with ICATCache() as icat:
                        experiment_details = icat.get_experiment_details(int(reference_number))
                except ICATConnectionException as excep:
                    render_error(request, str(excep))
            else:
                try:
                    with ICATCache(AUTH='uows', SESSION={'sessionid': request.session['sessionid']}) as icat:
                        experiment_details = icat.get_experiment_details(int(reference_number))
                except ICATConnectionException as excep:
                    render_error(request, str(excep))
        # pylint:disable=broad-except
        except Exception as icat_e:
            LOGGER.error(icat_e)
            experiment_details = {
                'reference_number': '',
                'start_date': '',
                'end_date': '',
                'title': '',
                'summary': '',
                'instrument': '',
                'pi': '',
            }
        context_dictionary = {
            'experiment': experiment,
            'runs_with_started_by': runs_with_started_by,
            'run_count': len(runs),
            'experiment_details': experiment_details,
            'data': data,
            'reduced_data': reduced_data,
        }
    # pylint:disable=broad-except
    except Exception as exception:
        LOGGER.error(exception)
        context_dictionary = {}

    return context_dictionary


@render_with('help.html')
# pylint:disable=redefined-builtin
def help(_):
    """
    Render help page
    Note: _ is replacing the passed in request parameter
    """
    return {}


@render_with('accessibility_statement.html')
# pylint:disable=redefined-builtin
def accessibility_statement(_):
    """
    Render accessibility statement page
    Note: _ is replacing the passed in request parameter
    """
    return {}


@render_with('admin/graph_home.html')
# pylint:disable=no-member
def graph_home(_):
    """
    Render graph page
    Note: _ is replacing the passed in request parameter
    """
    instruments = Instrument.objects.all()

    context_dictionary = {'instruments': instruments}

    return context_dictionary


@require_admin
@render_with('admin/graph_instrument.html')
# pylint:disable=no-member
def graph_instrument(request, instrument_name):
    """
    Render instrument specific graphing page
    """
    instrument = Instrument.objects.filter(name=instrument_name)
    if not instrument:
        return HttpResponseNotFound('<h1>Instrument not found</h1>')

    runs = (
        ReductionRun.objects.
        # Get the foreign key 'status' now. Otherwise many queries
        # made from load_runs which is very slow.
        select_related('status')
        # Only get these attributes, to speed it up.
        .only('status', 'started', 'finished', 'last_updated', 'created', 'run_number', 'run_description',
              'run_version').filter(instrument=instrument.first()).order_by('-created'))

    try:
        if 'last' in request.GET:
            runs = runs[:int(request.GET.get('last'))]
    except ValueError:
        # Non integer value entered as 'last' parameter.
        # Just show all runs.
        pass

    # Reverse list so graph displayed in correct order
    runs = runs[::-1]

    for run in runs:
        if run.started and run.finished:
            run.run_time = (run.finished - run.started).total_seconds()
        else:
            run.run_time = 0

    context_dictionary = {'runs': runs, 'instrument': instrument.first().name}

    return context_dictionary


@require_admin
@render_with('admin/stats.html')
# pylint:disable=no-member
def stats(_):
    """
    Render run statistics page
    Note: _ is replacing the passed in request parameter
    """
    statuses = []
    for status in Status.objects.all():
        status_count = (
            ReductionRun.objects.
            # Get the foreign key 'status' now.
            # Otherwise many queries made from load_runs which is very slow.
            select_related('status')
            # Only get these attributes, to speed it up.
            .only('status').filter(status__value=status.value).count())
        statuses.append({'name': status, 'count': status_count})

    context_dictionary = {
        'statuses': statuses,
    }

    return context_dictionary


def started_by_id_to_name(started_by_id=None):
    """
    Returns name of the user or team that submitted an autoreduction run given an ID number
    :param started_by_id: (int) The ID of the user who started the run, or a control code if not
     started by a user
    :return:
        If started by a valid user, returns the name either of the user in the format
         '[forename] [surname]'.
        If started automatically, returns "Autoreducton service".
        If started manually, returns "Development team".
        Otherwise, returns None.
    """
    if started_by_id is None or started_by_id < -1:
        return None

    if started_by_id == -1:
        return "Development team"

    if started_by_id == 0:
        return "Autoreduction service"

    try:
        user_record = User.objects.get(id=started_by_id)
        return f"{user_record.first_name} {user_record.last_name}"
    except ObjectDoesNotExist as exception:
        LOGGER.error(exception)
        return None



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Utility functions for the reduction viewer models

Note: This file has a large number of pylint disable as it was un unit tested
at the time of fixing the pylint issues. Once unit tested properly, these disables
should be able to be removed. Many are relating to imports
"""
import time

from autoreduce_db.reduction_viewer.models import ReductionRun
from autoreduce_utils.message.message import Message
from autoreduce_utils.clients.queue_client import QueueClient

from autoreduce_webapp.autoreduce_django.settings import FACILITY


class ReductionRunUtils:
    """
    Utilities for the ReductionRun model
    """
    @staticmethod
    def make_kwargs_from_runvariables(reduction_run, use_value=False) -> dict:
        """
        Given a reduction run and optionally use_value create a kwarg dict from its variables
        :param reduction_run: The ReductionRun
        :param use_value: Flag whether to use the variable values
        :return: kwarg dict of the variables
        """
        if reduction_run.run_variables.count() == 0:
            return {"standard_vars": {}, "advanced_vars": {}}

        return ReductionRunUtils.make_kwargs_from_variables(
            [runvar.variable for runvar in reduction_run.run_variables.all()], use_value)

    @staticmethod
    def make_kwargs_from_variables(variables, use_value=False) -> dict:
        """
        Given a list of variables, create a kwarg dict of the variables
        :param variables: The variables to make into a kwarg dict
        :param use_value: Flag whether to use the variable values
        :return: Kwarg dict of variables
        """
        standard_vars = {}
        advanced_vars = {}

        for variable in variables:
            if variable.is_advanced:
                advanced_vars[variable.name] = variable.value if use_value else variable
            else:
                standard_vars[variable.name] = variable.value if use_value else variable

        return {"standard_vars": standard_vars, "advanced_vars": advanced_vars}

    @staticmethod
    def send_retry_message(user_id: int, most_recent_run: ReductionRun, run_description: str, script_text: str,
                           new_script_arguments: dict, overwrite_previous_data: bool) -> None:
        """
        Creates & sends a retry message given the parameters

        :param user_id: The user submitting the run
        :param most_recent_run: The most recent run, used for common things across the two runs like
                                run number, instrument name, etc
        :param run_description: Description of the rerun
        :param script_text: The script that will NOT be used for this reduction, because of a known issue
                            https://github.com/ISISScientificComputing/autoreduce/issues/1115
        :param new_script_arguments: Dict of arguments that will be used for the reduction
        :param overwrite_previous_data: Whether to overwrite the previous data in the data location
        """
        message = Message(started_by=user_id,
                          description=run_description,
                          run_number=most_recent_run.run_number,
                          instrument=most_recent_run.instrument.name,
                          rb_number=most_recent_run.experiment.reference_number,
                          data=most_recent_run.data_location.first().file_path,
                          reduction_script=script_text,
                          reduction_arguments=new_script_arguments,
                          run_version=most_recent_run.run_version,
                          facility=FACILITY,
                          software=str(most_recent_run.software),
                          overwrite=overwrite_previous_data)
        MessagingUtils.send(message)

    @staticmethod
    def send_retry_message_same_args(user_id: int, most_recent_run: ReductionRun):
        """
        Sends a retry message using the parameters from the most_recent_run
        """
        ReductionRunUtils.send_retry_message(
            user_id, most_recent_run, "Re-run from the failed queue", most_recent_run.script,
            ReductionRunUtils.make_kwargs_from_runvariables(most_recent_run, use_value=True), most_recent_run.overwrite)


class ScriptUtils:
    """
    Utilities for the scripts field
    """
    @staticmethod
    def get_reduce_scripts(scripts):
        """
        Returns a tuple of (reduction script, reduction vars script),
        each one a string of the contents of the script, given a list of script objects.
        """
        script_out = None
        script_vars_out = None
        for script in scripts:
            if script.file_name == "reduce.py":
                script_out = script
            elif script.file_name == "reduce_vars.py":
                script_vars_out = script
        return script_out, script_vars_out

    def get_cache_scripts_modified(self, scripts):
        """
        :returns: The last time the scripts in the database were
        modified (in seconds since epoch).
        """
        script_modified = None
        script_vars_modified = None

        for script in scripts:
            if script.file_name == "reduce.py":
                script_modified = self._convert_time_from_string(str(script.created))
            elif script.file_name == "reduce_vars.py":
                script_vars_modified = self._convert_time_from_string(str(script.created))
        return script_modified, script_vars_modified

    @staticmethod
    def _convert_time_from_string(string_time):
        """
        :return: time as integer for epoch
        """
        time_format = "%Y-%m-%d %H:%M:%S"
        string_time = string_time[:string_time.find('+')]
        return int(time.mktime(time.strptime(string_time, time_format)))


class MessagingUtils:
    """
    Utilities for sending messages to ActiveMQ
    """
    @staticmethod
    def send(message):
        """ Sends message to ReductionPending (with the specified delay) """
        message_client = QueueClient()
        message_client.connect()

        message_client.send('/queue/DataReady', message, priority='1')
        message_client.disconnect()



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Custom parsing of user input
"""


def parse_user_run_numbers(user_input):
    """
    Returns an inclusive range of values based on the unsanitised user input as a string.
    If the input is malformed a syntax error is thrown, otherwise a single integer or list
    of integers are returned depending on the original input.
    :param user_input : The string representation of the user's input
    :return List or single value(s) based on user input
    :raises Syntax error if the users input contains malformed chars
    """
    allowed_non_numeric = ['-', ',']
    is_list = any(x in user_input for x in allowed_non_numeric)

    # Single value handling
    if not is_list:
        _check_input_is_numeric(user_input)
        return [int(user_input)]

    # List handling below
    _check_ranged_numeric_input(user_input)

    single_values = user_input.split(',')
    run_numbers = []
    for value in single_values:
        if '-' not in value:
            # Single value
            run_numbers.append(int(value))
            continue

        # Otherwise range separated values
        saturated_vals = _parse_range_input(value)
        if len(saturated_vals) == 2:
            run_numbers.extend(range(int(saturated_vals[0]), int(saturated_vals[1]) + 1))
        else:
            # Single negative value
            run_numbers.append(int(saturated_vals[0]))

    # Range is inclusive
    return run_numbers


def _check_input_is_numeric(str_input, extra_whitelisted_char_set=None):
    # Allow 0-9 to delimit the list plus any passed extra ones
    if not str_input:
        raise SyntaxError("No numeric digits were entered at all, or after a ',' or '-' character.")

    allowed_characters = set("1234567890-")
    if extra_whitelisted_char_set:
        allowed_characters |= set(extra_whitelisted_char_set)

    disallowed_char = next((char for char in str_input if char not in allowed_characters), None)
    if disallowed_char:
        raise SyntaxError("The character '{0}' was entered, " "which is not permitted".format(disallowed_char))

    return True


def _check_ranged_numeric_input(str_input):
    # We also allow dashes or commas for delimiting
    _check_input_is_numeric(str_input, '-,')

    # First split out all single values
    single_values = str_input.split(',')

    for value in single_values:
        if '-' not in value:
            # Single value
            _check_input_is_numeric(value)
            continue

        # Otherwise range separated values
        returned_ranges = _parse_range_input(value)

        if len(returned_ranges) > 2:
            # A range greater than 2 has been added (i.e. 3 - 5 - 7)
            raise SyntaxError("More than 2 values have been detected in {0}".format(str_input))

        for split_ranges in returned_ranges:
            _check_input_is_numeric(split_ranges)

    return True


def _parse_range_input(ranged_input):
    split_range = ranged_input.split('-')

    found_val = []

    next_val_represents_negative = False
    for value in split_range:
        if not value and not next_val_represents_negative:
            # The next value is probably negative
            next_val_represents_negative = True
            continue

        if not value and next_val_represents_negative:
            # We probably have an input like -3 - (-5) as: -3--5
            # consume this range separator
            continue

        # Check if the current value is negative
        found_val.append('-' + value if next_val_represents_negative else value)
        next_val_represents_negative = False

    return found_val



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Module for the run summary page model
"""
from typing import List

from django.urls.base import reverse
from selenium.webdriver.remote.webelement import WebElement
from autoreduce_webapp.selenium_tests.pages.component_mixins.footer_mixin import FooterMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.navbar_mixin import NavbarMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.rerun_form_mixin import \
    RerunFormMixin
from autoreduce_webapp.selenium_tests.pages.component_mixins.tour_mixin import TourMixin
from autoreduce_webapp.selenium_tests.pages.page import Page


class RunSummaryPage(Page, RerunFormMixin, NavbarMixin, FooterMixin, TourMixin):
    """
    Page model class for run summary page
    """
    def __init__(self, driver, instrument, run_number, version):
        super().__init__(driver)
        self.instrument = instrument
        self.run_number = run_number
        self.version = version

    def url_path(self) -> str:
        """
        Return the current URL of the page.
        :return: (str) the url path
        """
        return reverse("runs:summary",
                       kwargs={
                           "instrument_name": self.instrument,
                           "run_number": self.run_number,
                           "run_version": self.version
                       })

    @property
    def reduction_job_panel(self) -> WebElement:
        """Finds the run summary panel on the page."""
        return self.driver.find_element_by_id("reduction_job_panel")

    @property
    def rerun_form(self) -> WebElement:
        """Finds and returns the rerun form on the page"""
        return self.driver.find_element_by_id("rerun_form")

    @property
    def toggle_button(self) -> WebElement:
        """
        Finds and returns the toggle button for toggling the form on the page.
        """
        return self.driver.find_element_by_id("toggle_form")

    @property
    def reset_to_initial_values(self) -> WebElement:
        """
        Finds and returns the "Reset to original script and values" button
        """
        return self.driver.find_element_by_id("resetValues")

    @property
    def reset_to_current_values(self) -> WebElement:
        """
        Finds and returns the "Reset to values in the current reduce_vars script" button
        """
        return self.driver.find_element_by_id("currentScript")

    @property
    def warning_message(self) -> WebElement:
        """
        Finds and returns the "warning_message" box
        """
        return self.driver.find_element_by_id("warning_message")

    def run_description_text(self) -> str:
        """
        Finds and returns the text of the run_description field
        """
        return self.driver.find_element_by_id("run_description").text

    def started_by_text(self) -> str:
        """
        Finds and returns the text of the started_by field
        """
        return self.driver.find_element_by_id("started_by").text

    def status_text(self) -> str:
        """
        Finds and returns the text of the status field
        """
        return self.driver.find_element_by_id("status").text

    def instrument_text(self) -> str:
        """
        Finds and returns the text of the instrument field
        """
        return self.driver.find_element_by_id("instrument").text

    def rb_number_text(self) -> str:
        """
        Finds and returns the text of the rb_number field
        """
        return self.driver.find_element_by_id("rb_number").text

    def last_updated_text(self) -> str:
        """
        Finds and returns the text of the last_updated field
        """
        return self.driver.find_element_by_id("last_updated").text

    def reduction_host_text(self) -> WebElement:
        """
        Returns the reduction host text
        """
        return self.driver.find_element_by_id("reduction_host").text

    def images(self) -> List[WebElement]:
        """
        Returns all image elements on the page.
        """
        return self.driver.find_elements_by_tag_name("img")

    def plotly_plots(self) -> List[WebElement]:
        """
        Returns all image elements on the page.
        """
        return self.driver.find_elements_by_class_name("js-plotly-plot")



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Client for accessing the user office logon
"""
import logging

import suds
from suds.client import Client

# The below is a template on the repository
from autoreduce_webapp.autoreduce_django.settings import UOWS_URL

LOGGER = logging.getLogger("app")


class UOWSClient:
    """
    A client for interacting with the User Office Web Service
    """
    def __init__(self, **kwargs):
        url = kwargs.get("URL", UOWS_URL)
        self.client = Client(url)

    # Add the ability to use 'with'
    def __enter__(self):
        return self

    def __exit__(self, _, value, traceback):
        pass

    def check_session(self, session_id):
        """
        Checks if a session ID is still active and valid
        """
        try:
            return self.client.service.isTokenValid(session_id)
        except suds.WebFault:
            LOGGER.warning("Session ID is not valid: %s", session_id)
            return False

    def get_person(self, session_id):
        """
        Returns a dictionary containing basic person details
        for the user associated with the session id.
        Values include, first name, last name, email and unique usernumber.
        If session_id isn't valid, None is returned.
        """
        try:
            person = self.client.service.getPersonDetailsFromSessionId(session_id)
            if person:
                first_name = person.givenName
                if not first_name:
                    first_name = person.firstNameKnownAs
                trimmed_person = {
                    'first_name': first_name,
                    'last_name': person.familyName,
                    'email': person.email,
                    'usernumber': person.userNumber
                }
                return trimmed_person
        except suds.WebFault:
            LOGGER.warning("Session ID is not valid: %s", session_id)
        return None

    def logout(self, session_id):
        """
        Ends the session within the User Office Web Service.
        Note: This doesn't kill the local session.
        """
        try:
            self.client.service.logout(session_id)
        except suds.WebFault:
            LOGGER.warning("Failed to logout Session ID %s", session_id)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Register all the django models for reduction viewer
"""
from django.contrib import admin
from autoreduce_db.reduction_viewer.models import (Instrument, Experiment, Status, ReductionRun, DataLocation,
                                                   ReductionLocation, Setting, Notification)

admin.site.register(Instrument)
admin.site.register(Experiment)
admin.site.register(Status)
admin.site.register(ReductionRun)
admin.site.register(DataLocation)
admin.site.register(ReductionLocation)
admin.site.register(Setting)
admin.site.register(Notification)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Selenium tests for the runs summary page
"""

import os
import tempfile

from autoreduce_db.reduction_viewer.models import ReductionRun

from autoreduce_webapp.selenium_tests.pages.run_summary_page import RunSummaryPage
from autoreduce_webapp.selenium_tests.tests.base_tests import BaseTestCase


# pylint:disable=no-member
class TestRunSummaryPagePlots(BaseTestCase):
    """
    Test cases for the InstrumentSummary page when the Rerun form is NOT visible
    """

    fixtures = BaseTestCase.fixtures + ["one_run_plot"]

    def setUp(self) -> None:
        """
        Set up the instrument name and page
        """
        super().setUp()
        self.instrument_name = "TestInstrument"

        self.page = RunSummaryPage(self.driver, self.instrument_name, 99999, 0)
        self.run = ReductionRun.objects.first()

    def test_plot_files_png(self):
        """
        Test: PNG plot files are fetched and shown
        """
        # the plot files are expected to be in the reduction location, so we write them there for the test to work
        plot_files = [
            tempfile.NamedTemporaryFile(prefix="data_",
                                        suffix=".png",
                                        dir=self.run.reduction_location.first().file_path),
            tempfile.NamedTemporaryFile(prefix="data_",
                                        suffix=".png",
                                        dir=self.run.reduction_location.first().file_path)
        ]
        self.page.launch()

        # 1 is the logo, the other 2 are the plots
        images = self.page.images()
        assert len(images) == 3
        for img in images[1:]:
            alt_text = img.get_attribute("alt")
            assert "Plot image stored at" in alt_text
            assert any(os.path.basename(f.name) in alt_text for f in plot_files)

    def test_plot_files_json(self):
        """
        Test: JSON plot files are fetched and rendered by plotly
        """
        # the plot files are expected to be in the reduction location, so we write them there for the test to work
        plot_files = []

        for _ in range(2):
            # pylint:disable=consider-using-with
            tfile = tempfile.NamedTemporaryFile('w',
                                                prefix="data_",
                                                suffix=".json",
                                                dir=self.run.reduction_location.first().file_path)
            tfile.write("""{"data": [{"type": "bar","x": [1,2,3],"y": [1,3,2]}]}""")
            tfile.flush()
            plot_files.append(tfile)

        self.page.launch()

        plots = self.page.plotly_plots()
        assert len(plots) == 2
        for plot in plots:
            assert any(plot.get_attribute("id") in file.name for file in plot_files)

    def test_plot_files_mix(self):
        """Test that both static and interactive plots are rendered"""
        plot_files = [
            tempfile.NamedTemporaryFile(prefix="data_",
                                        suffix=".png",
                                        dir=self.run.reduction_location.first().file_path),
            tempfile.NamedTemporaryFile('w',
                                        prefix="data_",
                                        suffix=".json",
                                        dir=self.run.reduction_location.first().file_path)
        ]
        # write the interactive plot data
        plot_files[1].write("""{"data": [{"type": "bar","x": [1,2,3],"y": [1,3,2]}]}""")
        plot_files[1].flush()

        self.page.launch()

        images = self.page.images()
        # 1 is the logo, the other is the static plot
        assert len(images) == 2

        img = images[1]
        alt_text = img.get_attribute("alt")
        assert "Plot image stored at" in alt_text
        assert any(os.path.basename(f.name) in alt_text for f in plot_files)

        plots = self.page.plotly_plots()
        assert len(plots) == 1
        assert plots[0].get_attribute("id") in plot_files[1].name



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #

from django.urls import path
from autoreduce_webapp.instrument.views import runs, variables, pause

app_name = "instrument"

urlpatterns = [
    path('<str:instrument>/submit_runs/', runs.submit_runs, name='submit_runs'),
    path('<str:instrument>/configure_new_runs/', runs.configure_new_runs, name='variables'),
    path('<str:instrument>/configure_new_runs/<int:start>/', runs.configure_new_runs, name='variables'),
    path('<str:instrument>/configure_new_runs/<int:start>/<int:end>/', runs.configure_new_runs, name='variables'),
    path('<str:instrument>/variables_summary/', variables.instrument_variables_summary, name='variables_summary'),
    path('<str:instrument>/variables/<int:start>/<int:end>/delete',
         variables.delete_instrument_variables,
         name='delete_variables'),
    path('<str:instrument>/variables/experiment/<int:experiment_reference>/',
         runs.configure_new_runs,
         name='variables_by_experiment'),
    path('<str:instrument>/variables/experiment/<int:experiment_reference>/delete/',
         variables.delete_instrument_variables,
         name='delete_variables_by_experiment'),
    path('<str:instrument>/default_variables/', variables.current_default_variables, name='current_default_variables'),
    path('<str:instrument>/pause/', pause.instrument_pause, name='pause'),
]



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Selenium tests for the runs summary page
"""

from django.urls import reverse
from autoreduce_db.reduction_viewer.models import ReductionRun
from autoreduce_qp.systemtests.utils.data_archive import DataArchive
from selenium.webdriver.support.wait import WebDriverWait

from autoreduce_webapp.selenium_tests.pages.run_summary_page import RunSummaryPage
from autoreduce_webapp.selenium_tests.tests.base_tests import BaseTestCase, FooterTestMixin, NavbarTestMixin


# pylint:disable=no-member
class TestRunSummaryPage(NavbarTestMixin, BaseTestCase, FooterTestMixin):
    """
    Test cases for the InstrumentSummary page when the Rerun form is NOT visible
    """

    fixtures = BaseTestCase.fixtures + ["run_with_one_variable"]

    @classmethod
    def setUpClass(cls):
        """Sets up Dataarchive with scripts and sets instrument for all test cases"""
        super().setUpClass()
        cls.instrument_name = "TestInstrument"
        cls.data_archive = DataArchive([cls.instrument_name], 21, 21)
        cls.data_archive.create()
        cls.data_archive.add_reduce_vars_script(cls.instrument_name,
                                                """standard_vars={"variable1":"test_variable_value_123"}""")

    @classmethod
    def tearDownClass(cls) -> None:
        """Destroys created data archive"""
        cls.data_archive.delete()
        super().tearDownClass()

    def setUp(self) -> None:
        """Sets up RunSummaryPage before each test case"""
        super().setUp()
        self.page = RunSummaryPage(self.driver, self.instrument_name, 99999, 0)
        self.page.launch()

    def test_reduction_job_panel_displayed(self):
        """Tests that the reduction job panel is showing the right things"""
        # only one run in the fixture, get it for assertions
        run = ReductionRun.objects.first()
        assert self.page.reduction_job_panel.is_displayed()
        assert self.page.run_description_text() == f"Run description: {run.run_description}"
        # because it's started_by: -1, determined in `started_by_id_to_name`
        assert self.page.started_by_text() == "Started by: Development team"
        assert self.page.status_text() == "Status: Processing"
        assert self.page.instrument_text() == f"Instrument: {run.instrument.name}"
        assert self.page.rb_number_text() == f"RB Number: {run.experiment.reference_number}"
        assert self.page.last_updated_text() == "Last Updated: 19 Oct 2020, 6:35 p.m."
        assert self.page.reduction_host_text() == "Host: test-host-123"

    def test_reduction_job_panel_reset_to_values_first_used_for_run(self):
        """Test that the button to reset the variables to the values first used for the run works"""
        self.page.toggle_button.click()
        self.page.variable1_field = "the new value in the field"

        self.page.reset_to_initial_values.click()

        # need to re-query the driver because resetting replaces the elements
        assert self.page.variable1_field.get_attribute("value") == "value1"

    def test_reduction_job_panel_reset_to_current_reduce_vars(self):
        """Test that the button to reset the variables to the values from the reduce_vars script works"""
        self.page.toggle_button.click()
        self.page.variable1_field = "the new value in the field"

        self.page.reset_to_current_values.click()

        # need to re-query the driver because resetting replaces the elements
        assert self.page.variable1_field.get_attribute("value") == "test_variable_value_123"

    def test_rerun_form(self):
        """
        Test: Rerun form shows contents from Variable in database (from the fixture) and not reduce_vars.py
        """
        rerun_form = self.page.rerun_form
        assert not rerun_form.is_displayed()
        self.page.toggle_button.click()
        assert rerun_form.is_displayed()
        assert rerun_form.find_element_by_id("var-standard-variable1").get_attribute("value") == "value1"
        labels = rerun_form.find_elements_by_tag_name("label")

        WebDriverWait(self.driver, 10).until(lambda _: labels[0].text == "Re-run description")
        WebDriverWait(self.driver, 10).until(lambda _: labels[1].text == "variable1")

    def test_back_to_instruments_goes_back(self):
        """
        Test: Clicking back goes back to the instrument
        """
        back = self.page.cancel_button
        assert back.is_displayed()
        assert back.text == f"Back to {self.instrument_name} runs"
        back.click()
        assert reverse("runs:list", kwargs={"instrument": self.instrument_name}) in self.driver.current_url



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #

from django.conf import settings
from django.conf.urls import include
from django.contrib import admin
from django.urls import path, register_converter
from autoreduce_webapp.instrument.views import runs
from autoreduce_webapp.reduction_viewer import views as reduction_viewer_views

# pylint: disable=invalid-name
handler400 = 'autoreduce_webapp.autoreduce_django.views.handler400'
handler403 = 'autoreduce_webapp.autoreduce_django.views.handler403'
handler404 = 'autoreduce_webapp.autoreduce_django.views.handler404'
handler500 = 'autoreduce_webapp.autoreduce_django.views.handler500'


class NegativeIntConverter:
    regex = r'-?\d+'

    @staticmethod
    def to_python(value):
        """
        Return the value as a Python object
        """
        return int(value)

    @staticmethod
    def to_url(value):
        """
        Return the value as a URL string
        """
        return '%d' % value


register_converter(NegativeIntConverter, 'negint')

urlpatterns = [
    # ===========================MISC================================= #
    path('', reduction_viewer_views.index, name='index'),
    path('admin/', admin.site.urls),
    path('logout/', reduction_viewer_views.logout, name='logout'),
    path('help/', reduction_viewer_views.help, name='help'),
    path('accessibility_statement/', reduction_viewer_views.accessibility_statement, name='accessibility_statement'),

    # ===========================RUNS================================= #
    path('overview/', reduction_viewer_views.overview, name='overview'),
    path('runs/', include('reduction_viewer.urls')),
    path('runs/<str:instrument>/confirmation/', runs.run_confirmation, name='run_confirmation'),

    # ===========================INSTRUMENT========================== #
    path('instrument/', include('instrument.urls')),

    # ===========================EXPERIMENT========================== #
    path('experiment/<negint:reference_number>/', reduction_viewer_views.experiment_summary, name='experiment_summary'),

    # ===========================SCRIPTS============================= #
    path('graph/', reduction_viewer_views.graph_home, name="graph"),
    path('graph/<str:instrument_name>', reduction_viewer_views.graph_instrument, name="graph_instrument"),
    path('stats/', reduction_viewer_views.stats, name="stats"),
]

if settings.DEBUG:
    try:
        import debug_toolbar

        urlpatterns = [
            path('__debug__/', include(debug_toolbar.urls)),
        ] + urlpatterns
    except ModuleNotFoundError:
        # debug_toolbar not installed - just run without it
        pass



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #

from django.urls import path
from autoreduce_webapp.reduction_viewer import views

app_name = "runs"

urlpatterns = [
    path('queue/', views.run_queue, name='queue'),
    path('failed/', views.fail_queue, name='failed'),
    path('<str:instrument>/', views.runs_list, name='list'),
    path('<str:instrument_name>/<int:run_number>/', views.run_summary, name='summary'),
    path('<str:instrument_name>/<int:run_number>/<int:run_version>/', views.run_summary, name='summary'),
]



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Selenium tests for the runs summary page
"""

from autoreduce_qp.systemtests.utils.data_archive import DataArchive

from autoreduce_webapp.selenium_tests.pages.runs_list_page import RunsListPage
from autoreduce_webapp.selenium_tests.tests.base_tests import NavbarTestMixin, BaseTestCase, FooterTestMixin, \
    AccessibilityTestMixin


class TestRunsListPage(NavbarTestMixin, BaseTestCase, FooterTestMixin, AccessibilityTestMixin):
    """
    Test cases for the InstrumentSummary page
    """

    fixtures = BaseTestCase.fixtures + ["test_runs_list_page"]

    def setUp(self) -> None:
        """
        Sets up the InstrumentSummaryPage object
        """
        super().setUp()
        self.instrument_name = "TestInstrument"
        self.page = RunsListPage(self.driver, self.instrument_name)

    def test_reduction_run_displayed(self):
        """
        Test: Reduction run is displayed
        When: The run exists in the database
        """
        runs = self.page.launch().get_run_numbers_from_table()
        assert "99999" in runs

    def test_alert_message_when_missing_reduce_vars(self):
        """
        Test that the correct message is shown when the reduce_vars.py file is missing
        """
        self.page.launch()
        expected = "The buttons above have been disabled because reduce_vars.py is missing for this instrument."
        assert self.page.alert_message_text() == expected

    def test_alert_message_when_reduce_vars_has_error(self):
        """
        Test that the correct message is shown when the reduce_vars.py has an error in it
        """
        data_archive = DataArchive([self.instrument_name], 21, 21)
        data_archive.create()

        # add a reduce_vars script with a syntax error -> a missing " after value1
        data_archive.add_reduce_vars_script(self.instrument_name, """standard_vars={"variable1":"value1}""")

        self.page.launch()
        expected = "The buttons above have been disabled because reduce_vars.py has an import or syntax error."
        assert self.page.alert_message_text() == expected

        data_archive.delete()



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Utility functions for the view of django models
"""
import functools
import logging
import os

from autoreduce_db.reduction_viewer.models import Instrument
from autoreduce_qp.queue_processor.reduction.service import ReductionScript

from autoreduce_webapp.autoreduce_django.settings import DATA_ANALYSIS_BASE_URL

LOGGER = logging.getLogger("app")


def deactivate_invalid_instruments(func):
    """
    Deactivate instruments if they are invalid
    """
    @functools.wraps(func)
    def request_processor(request, *args, **kws):
        """
        Function decorator that checks the reduction script for all active instruments
        and deactivates any that cannot be found
        """
        instruments = Instrument.objects.all()
        for instrument in instruments:
            script_path = ReductionScript(instrument.name)
            if instrument.is_active != script_path.exists():
                instrument.is_active = script_path.exists()
                instrument.save(update_fields=['is_active'])

        return func(request, *args, **kws)

    return request_processor


def get_interactive_plot_data(plot_locations):
    """
    Get the data for the interactive plots from the saved JSON files
    """
    # get only JSON files
    json_files = [location for location in plot_locations if location.endswith(".json")]

    output = {}

    for filepath in json_files:
        name = os.path.basename(filepath)
        with open(filepath, 'r') as file:
            data = file.read()
        output[name] = data
    return output


def make_data_analysis_url(reduction_location: str):
    """Makes a URL for the data.analysis website that will open the location of the data"""
    if "/instrument/" in reduction_location:
        return DATA_ANALYSIS_BASE_URL + reduction_location.split("/instrument/")[1]
    else:
        return ""



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Renders the time difference between to given times
"""
from django.template import Library, Node, Variable, TemplateSyntaxError
from django.template.defaultfilters import pluralize

from autoreduce_webapp.autoreduce_django.templatetags.common_helpers import get_var

# pylint:disable=invalid-name
register = Library()


class NaturalTimeDifferenceNode(Node):
    """
    Class for computing and rendering time differences
    """
    def __init__(self, start, end):
        self.start = Variable(start)
        self.end = Variable(end)

    def render(self, context):
        """
        Render the response
        """
        start = get_var(self.start, context)
        end = get_var(self.end, context)
        delta = end - start
        days = delta.days
        hours, remainder = divmod(delta.seconds, 3600)
        minutes, remainder = divmod(remainder, 60)
        seconds = remainder
        human_delta = ''
        if days > 0:
            if human_delta:
                human_delta += ', '
            human_delta += '%i day%s' % (days, pluralize(days))
        if hours > 0:
            if human_delta:
                human_delta += ', '
            human_delta += '%i hour%s' % (hours, pluralize(hours))
        if minutes > 0:
            if human_delta:
                human_delta += ', '
            human_delta += '%i minute%s' % (minutes, pluralize(minutes))
        if seconds > 0:
            if human_delta:
                human_delta += ', '
            human_delta += '%i second%s' % (seconds, pluralize(seconds))
        return human_delta


def natural_time_difference(_, token):
    """
    Return NaturalTimeDifference Node
    """
    args = token.split_contents()[1:]
    if len(args) != 2:
        raise TemplateSyntaxError('%r tag requires two datetimes.' % token.contents.split()[0])
    return NaturalTimeDifferenceNode(*args)


register.tag('natural_time_difference', natural_time_difference)



from django.urls import reverse
from autoreduce_qp.systemtests.utils.data_archive import DataArchive

from autoreduce_webapp.selenium_tests.pages.rerun_jobs_page import RerunJobsPage
from autoreduce_webapp.selenium_tests.tests.base_tests import (NavbarTestMixin, BaseTestCase, FooterTestMixin,
                                                               AccessibilityTestMixin)


# pylint:disable=no-member
class TestRerunJobsPage(NavbarTestMixin, BaseTestCase, FooterTestMixin, AccessibilityTestMixin):
    """
    Test cases for the InstrumentSummary page when the Rerun form is NOT visible
    """

    fixtures = BaseTestCase.fixtures + ["run_with_one_variable"]

    accessibility_test_ignore_rules = {
        # https://github.com/ISISScientificComputing/autoreduce/issues/1267
        "duplicate-id-aria": "input",
    }

    @classmethod
    def setUpClass(cls):
        """Sets up DataArchive for all tests and sets instrument for all tests"""
        super().setUpClass()
        cls.instrument_name = "TestInstrument"
        cls.data_archive = DataArchive([cls.instrument_name], 21, 21)
        cls.data_archive.create()
        cls.data_archive.add_reduction_script(cls.instrument_name, """print('some text')""")
        cls.data_archive.add_reduce_vars_script(cls.instrument_name,
                                                """standard_vars={"variable1":"test_variable_value_123"}""")

    @classmethod
    def tearDownClass(cls) -> None:
        """Destroys created DataArchive"""
        cls.data_archive.delete()
        super().tearDownClass()

    def setUp(self) -> None:
        """Set up RerunJobsPage before each test case"""
        super().setUp()
        self.page = RerunJobsPage(self.driver, self.instrument_name)
        self.page.launch()

    def test_cancel_goes_back_to_runs_list(self):
        """Tests: Clicking canel button returns the runs list page"""
        self.page.cancel_button.click()
        assert reverse("runs:list", kwargs={"instrument": self.instrument_name}) in self.driver.current_url

    def test_reset_values_does_reset_the_values(self):
        """Test that the button to reset the variables to the values from the reduce_vars script works"""
        self.page.variable1_field = "the new value in the field"
        self.page.reset_to_current_values.click()

        # need to re-query the driver because resetting replaces the elements
        var_field = self.page.variable1_field
        assert var_field.get_attribute("value") == "test_variable_value_123"



# pylint:skip-file
"""
Wrapper for the functionality for various installation and project setup commands
see:
    `python setup.py help`
for more details
"""
from setuptools import setup, find_packages

setup(name="autoreduce_webapp",
      version="21.1.3",
      description="The frontend of the ISIS Autoreduction service",
      author="ISIS Autoreduction Team",
      url="https://github.com/ISISScientificComputing/autoreduce-frontend/",
      install_requires=[
          "autoreduce_utils==0.1.3", "autoreduce_db==0.1.4", "autoreduce_qp==21.1.3", "Django==3.2.2",
          "django_extensions==3.1.3", "django-user-agents==0.4.0"
      ],
      packages=find_packages(),
      package_data={"": ["templates/*.*", "templates/**/*.*", "static/**/*.*", "**/*.json"]},
      entry_points={"console_scripts": ["autoreduce-webapp-manage = autoreduce_webapp.manage:main"]})



"""
Custom Paginator for the web application and associated Pages that are used to model each
individual page in the pagination.

This works in a similar way to the standard Paginator for Django but gives the added benefit of
easier display names, as well as setting a maximum number of pages to display in the pagination

This is currently used in the runs_list.html page
"""

import math


class PageLimitException(Exception):
    """ Custom Exception class to catch when page record limit is reached"""


class CustomPaginator:
    """
    CustomPaginator to allow for more complex functionality to be used without doing heavy
    lifting in the django template code
    """
    def __init__(self, page_type, query_set, items_per_page, page_tolerance, current_page):
        """
        :param query_set: All data to show
        :param items_per_page: Number of items per page
        :param page_tolerance: Number of pages to show
        """
        self.page_type = page_type
        self.query_set = query_set
        self.items_per_page = int(items_per_page)
        self.page_tolerance = int(page_tolerance)
        self.current_page_index = int(current_page)
        self.next_page_index = 0
        self.has_next = False
        self.has_previous = False
        self.previous_page_index = 0
        self.page_list = []
        self.display_list = []
        self._validate_current_page()
        self._construct_pagination()
        self._set_next_and_previous()
        if self.page_list:
            self.current_page = self.page_list[self.current_page_index - 1]
        self._create_display_list()

    def _validate_current_page(self):
        """
        Ensure that the current page specified is valid
        Update the page to min/max if outside of expected range
        """
        page_count = int(math.ceil(float(len(self.query_set)) / float(self.items_per_page)))
        self.current_page_index = max(1, min(self.current_page_index, page_count))

    def _construct_pagination(self):
        """
        Construct the pages, populate them with records
        and add them to the self.page_list variable.
        """
        if abs(self.current_page_index - 1) <= self.page_tolerance:
            current_page = self._create_page(self.page_type, 1, self.items_per_page, True)
        else:
            current_page = self._create_page(self.page_type, 1, self.items_per_page, False)

        for record in self.query_set:
            try:
                current_page.add_record(record)
            except PageLimitException:
                self.page_list.append(current_page)
                next_page_index = current_page.number + 1
                if abs(self.current_page_index - next_page_index) <= self.page_tolerance:
                    current_page = self._create_page(self.page_type, next_page_index, self.items_per_page, True)
                else:
                    current_page = self._create_page(self.page_type, next_page_index, self.items_per_page, False)
                # Make sure we add the record to the new page we created
                current_page.add_record(record)
        current_page.set_start_and_end()
        self.page_list.append(current_page)

    @staticmethod
    def _create_page(page_type, page_number, max_number_of_items, visible):
        if page_type.lower() == 'run':
            return RunPage(page_number, max_number_of_items, visible)
        if page_type.lower() == 'date':
            return DatePage(page_number, max_number_of_items, visible)
        # This should never be triggered, but let's default to Run display if there is a problem
        return 'run'

    def _create_display_list(self):
        """
        A compact version of the pagination table that is used for display purposes
        """
        for page in self.page_list:
            if page.is_visible:
                self.display_list.append(page)
            else:
                if len(self.display_list) >= 1 and self.display_list[-1] == "...":
                    continue
                self.display_list.append("...")

    def _set_next_and_previous(self):
        """
        Sets class variables to describe what page is before and after the current page
        """
        if self.current_page_index < len(self.page_list):
            self.next_page_index = self.current_page_index + 1
            self.has_next = True
        if self.current_page_index > 1:
            self.previous_page_index = self.current_page_index - 1
            self.has_previous = True


class CustomPage:
    """
    Generic CustomPage class to model a single page in the pagination
    """
    def __init__(self, number, max_limit, is_visible):
        """
        :param number: The page number
        :param max_limit: The most amount of data items allowed on a page
        :param is_visible: Should this page be shown in the pagination?
        """
        self.records = []
        self.is_visible = is_visible
        self.number = number
        self.max_limit = max_limit
        self.display_name = None
        self.start = 0
        self.end = 0

    def set_start_and_end(self):
        """ Implemented by child class """

    def add_record(self, record):
        """
        Adds a record to the page, if the page has reached the limit that it can hold,
        set the start and end run number variables and throw an exception to be caught
        """
        if len(self.records) >= self.max_limit:
            self.set_start_and_end()
            raise PageLimitException("Unable to add record to page. Max number of records reached")
        self.records.append(record)


class RunPage(CustomPage):
    """
    Specific implementation for rendering the pagination if sorting by Run Number
    """
    def set_start_and_end(self):
        """
        Custom function to set the start and end variables and
        construct the display name for the page
        """
        self.start = self.records[0].run_number
        self.end = self.records[-1].run_number
        if self.start == self.end:
            self.display_name = "{}".format(self.start)
        else:
            self.display_name = "{} - {}".format(self.start, self.end)


class DatePage(CustomPage):
    """
    Specific implementation for rendering the pagination if sorting by Date
    """
    def set_start_and_end(self):
        """
        Custom function to set the start and end variables and
        construct the display name for the page
        """
        self.start = self.records[0].last_updated
        self.end = self.records[-1].last_updated

        if self.start == self.end:
            self.display_name = "{}/{}/{}".format(self.start.day, self.start.month, self.start.year)
        else:
            self.display_name = "{}/{}/{} - {}/{}/{}".format(self.start.day, self.start.month, self.start.year,
                                                             self.end.day, self.end.month, self.end.year)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Initialise admin pages
"""
from django.contrib import admin

from autoreduce_webapp.autoreduce_django.models import UserCache, InstrumentCache, ExperimentCache

admin.site.register(UserCache)
admin.site.register(InstrumentCache)
admin.site.register(ExperimentCache)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Register models variables here
"""
from django.contrib import admin
from autoreduce_db.instrument.models import InstrumentVariable, RunVariable

# Register your models here.
admin.site.register(InstrumentVariable)
admin.site.register(RunVariable)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Selenium tests for the runs summary page
"""

from django.urls import reverse
from autoreduce_webapp.selenium_tests.pages.run_summary_page import RunSummaryPage
from autoreduce_webapp.selenium_tests.tests.base_tests import (BaseTestCase, FooterTestMixin, NavbarTestMixin,
                                                               AccessibilityTestMixin)
from autoreduce_webapp.selenium_tests.utils import submit_and_wait_for_result

from autoreduce_webapp.selenium_tests.utils import setup_external_services


class TestRunSummaryPageIntegration(BaseTestCase, FooterTestMixin, NavbarTestMixin, AccessibilityTestMixin):
    """
    Test cases for the InstrumentSummary page when the Rerun form is NOT visible
    """

    fixtures = BaseTestCase.fixtures + ["run_with_one_variable"]

    accessibility_test_ignore_rules = {
        # https://github.com/ISISScientificComputing/autoreduce/issues/1267
        # https://github.com/ISISScientificComputing/autoreduce/issues/1268
        "duplicate-id-aria": "input, #run_description",
    }

    @classmethod
    def setUpClass(cls):
        """ Start all external services """
        super().setUpClass()
        cls.instrument_name = "TestInstrument"
        cls.rb_number = 1234567
        cls.run_number = 99999
        cls.data_archive, cls.queue_client, cls.listener = setup_external_services(cls.instrument_name, 21, 21)
        cls.data_archive.add_reduction_script(cls.instrument_name, """print('some text')""")
        cls.data_archive.add_reduce_vars_script(cls.instrument_name,
                                                """standard_vars={"variable1":"test_variable_value_123"}""")

    @classmethod
    def tearDownClass(cls) -> None:
        """Stop all external services"""
        cls.queue_client.disconnect()
        cls.data_archive.delete()
        super().tearDownClass()

    def setUp(self) -> None:
        """Sets up the RunSummaryPage and shows the rerun panel before each test case"""
        super().setUp()
        self.page = RunSummaryPage(self.driver, self.instrument_name, 99999, 0)
        self.page.launch()
        # clicks the toggle to show the rerun panel, otherwise the buttons in the form are non interactive
        self.page.toggle_button.click()

    def test_submit_rerun_same_variables(self):
        """
        Test: Just opening the submit page and clicking rerun
        """
        result = submit_and_wait_for_result(self)
        assert len(result) == 2

        assert result[0].run_version == 0
        assert result[1].run_version == 1

        for run0_var, run1_var in zip(result[0].run_variables.all(), result[1].run_variables.all()):
            assert run0_var.variable == run1_var.variable

    def test_submit_rerun_changed_variable_arbitrary_value(self):
        """
        Test: Open submit page, change a variable, submit the run
        """
        # change the value of the variable field
        self.page.variable1_field = "the new value in the field"

        result = submit_and_wait_for_result(self)
        assert len(result) == 2

        assert result[0].run_version == 0
        assert result[1].run_version == 1

        for run0_var, run1_var in zip(result[0].run_variables.all(), result[1].run_variables.all()):
            # the value of the variable has been overwritten because it's the same run number
            assert run0_var.variable == run1_var.variable

        assert result[1].run_variables.first().variable.value == "the new value in the field"

    def test_submit_rerun_after_clicking_reset_initial(self):
        """
        Test: Submitting a run after changing the value and then clicking reset to initial values
        will correctly use the initial values
        """
        # change the value of the variable field
        self.page.variable1_field = "the new value in the field"

        self.page.reset_to_initial_values.click()
        result = submit_and_wait_for_result(self)
        assert len(result) == 2

        assert result[0].run_version == 0
        assert result[1].run_version == 1

        for run0_var, run1_var in zip(result[0].run_variables.all(), result[1].run_variables.all()):
            # the value of the variable has been overwritten because it's the same run number
            assert run0_var.variable == run1_var.variable

        assert result[1].run_variables.first().variable.value == "value1"

    def test_submit_rerun_after_clicking_reset_current_script(self):
        """
        Test: Submitting a run after clicking the reset to current script uses the values saved in the current script
        """
        self.page.reset_to_current_values.click()
        result = submit_and_wait_for_result(self)

        assert len(result) == 2

        assert result[0].run_version == 0
        assert result[1].run_version == 1

        for run0_var, run1_var in zip(result[0].run_variables.all(), result[1].run_variables.all()):
            # the value of the variable has been overwritten because it's the same run number
            assert run0_var.variable == run1_var.variable

        assert result[1].run_variables.first().variable.value == "test_variable_value_123"

    def test_submit_confirm_page(self):
        """
        Test: Submitting a run leads to the correct page
        """
        result = submit_and_wait_for_result(self)
        expected_url = reverse("run_confirmation", kwargs={"instrument": self.instrument_name})
        assert expected_url in self.driver.current_url
        # wait until the message processing is complete before ending the test
        # otherwise the message handling can pollute the DB state for the next test
        assert len(result) == 2
        # check that the error is because of missing Mantid
        # if this fails then something else in the reduction caused an error instead!
        assert "Mantid" in result[1].admin_log



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #

import logging
from itertools import chain

from autoreduce_db.instrument.models import InstrumentVariable
from autoreduce_db.reduction_viewer.models import Instrument, ReductionRun, Status
from autoreduce_qp.queue_processor.instrument_variable_utils import InstrumentVariablesUtils
from autoreduce_qp.queue_processor.reduction.service import ReductionScript
from autoreduce_qp.queue_processor.variable_utils import VariableUtils
from django.db.models.query import QuerySet
from django.shortcuts import redirect

from autoreduce_webapp.autoreduce_django.view_utils import (check_permissions, login_and_uows_valid, render_with)
from autoreduce_webapp.reduction_viewer.utils import ReductionRunUtils
from autoreduce_webapp.utilities import input_processing

LOGGER = logging.getLogger("app")


# pylint:disable=inconsistent-return-statements
@login_and_uows_valid
@check_permissions
@render_with('submit_runs.html')
def submit_runs(request, instrument=None):
    """
    Handles run submission request
    """
    LOGGER.info('Submitting runs')
    # pylint:disable=no-member
    instrument = Instrument.objects.prefetch_related('reduction_runs').get(name=instrument)
    if request.method == 'GET':
        processing_status = Status.get_processing()
        queued_status = Status.get_queued()

        # pylint:disable=no-member
        runs_for_instrument = instrument.reduction_runs.all()
        last_run = instrument.get_last_for_rerun(runs_for_instrument)

        kwargs = ReductionRunUtils.make_kwargs_from_runvariables(last_run)
        standard_vars = kwargs["standard_vars"]
        advanced_vars = kwargs["advanced_vars"]

        try:
            current_variables = VariableUtils.get_default_variables(instrument)
        except (FileNotFoundError, ImportError, SyntaxError) as err:
            return {"message": str(err)}

        current_standard_variables = current_variables["standard_vars"]
        current_advanced_variables = current_variables["advanced_vars"]
        # pylint:disable=no-member
        context_dictionary = {
            'instrument': instrument,
            'last_instrument_run': last_run,
            'processing': runs_for_instrument.filter(status=processing_status),
            'queued': runs_for_instrument.filter(status=queued_status),
            'standard_variables': standard_vars,
            'advanced_variables': advanced_vars,
            'current_standard_variables': current_standard_variables,
            'current_advanced_variables': current_advanced_variables,
        }

        return context_dictionary


# pylint:disable=too-many-return-statements,too-many-branches,too-many-statements,too-many-locals
@login_and_uows_valid
@check_permissions
@render_with('run_confirmation.html')
def run_confirmation(request, instrument: str):
    """
    Handles request for user to confirm re-run
    """
    range_string = request.POST.get('run_range')
    run_description = request.POST.get('run_description')

    # pylint:disable=no-member
    queue_count = ReductionRun.objects.filter(instrument__name=instrument, status=Status.get_queued()).count()
    context_dictionary = {
        # list stores (run_number, run_version)
        'runs': [],
        'variables': None,
        'queued': queue_count,
        'instrument_name': instrument,
        'run_description': run_description
    }

    try:
        run_numbers = input_processing.parse_user_run_numbers(range_string)
    except SyntaxError as exception:
        context_dictionary['error'] = exception.msg
        return context_dictionary

    if not run_numbers:
        context_dictionary['error'] = f"Could not correctly parse range input {range_string}"
        return context_dictionary

    # Determine user level to set a maximum limit to the number of runs that can be re-queued
    if request.user.is_superuser:
        max_runs = 500
    elif request.user.is_staff:
        max_runs = 50
    else:
        max_runs = 20

    if len(run_numbers) > max_runs:
        context_dictionary["error"] = "{0} runs were requested, but only {1} runs can be " \
                                      "queued at a time".format(len(run_numbers), max_runs)
        return context_dictionary

    related_runs: QuerySet[ReductionRun] = ReductionRun.objects.filter(instrument__name=instrument,
                                                                       run_number__in=run_numbers)
    # Check that RB numbers are the same for the range entered
    # pylint:disable=no-member
    rb_number = related_runs.values_list('experiment__reference_number', flat=True).distinct()
    if len(rb_number) > 1:
        context_dictionary['error'] = 'Runs span multiple experiment numbers ' \
                                      '(' + ','.join(str(i) for i in rb_number) + ')' \
                                      ' please select a different range.'
        return context_dictionary

    try:
        script_text = InstrumentVariablesUtils.get_current_script_text(instrument)
        default_variables = VariableUtils.get_default_variables(instrument)
    except (FileNotFoundError, ImportError, SyntaxError) as err:
        context_dictionary['error'] = err
        return context_dictionary

    try:
        new_script_arguments = make_reduction_arguments(request.POST.items(), default_variables)
        context_dictionary['variables'] = new_script_arguments
    except ValueError as err:
        context_dictionary['error'] = err
        return context_dictionary

    for run_number in run_numbers:
        matching_previous_runs = related_runs.filter(run_number=run_number).order_by('-run_version')
        run_suitable, reason = find_reason_to_avoid_re_run(matching_previous_runs, run_number)
        if not run_suitable:
            context_dictionary['error'] = reason
            break

        # run_description gets stored in run_description in the ReductionRun object
        max_run_description_length = ReductionRun._meta.get_field('run_description').max_length
        if len(run_description) > max_run_description_length:
            context_dictionary["error"] = "The description contains {0} characters, " \
                                          "a maximum of {1} are allowed".\
                format(len(run_description), max_run_description_length)
            return context_dictionary

        most_recent_run: ReductionRun = matching_previous_runs.first()
        # User can choose whether to overwrite with the re-run or create new data
        ReductionRunUtils.send_retry_message(request.user.id, most_recent_run, run_description, script_text,
                                             new_script_arguments, False)
        # list stores (run_number, run_version)
        context_dictionary["runs"].append((run_number, most_recent_run.run_version + 1))

    return context_dictionary


def find_reason_to_avoid_re_run(matching_previous_runs, run_number):
    """
    Check whether the most recent run exists
    """
    most_recent_run = matching_previous_runs.first()

    # Check old run exists - if it doesn't exist there's nothing to re-run!
    if most_recent_run is None:
        return False, f"Run number {run_number} hasn't been ran by autoreduction yet."

    # Prevent multiple queueings of the same re-run
    queued_runs = matching_previous_runs.filter(status=Status.get_queued()).first()
    if queued_runs is not None:
        return False, f"Run number {queued_runs.run_number} is already queued to run"

    return True, ""


def make_reduction_arguments(post_arguments, default_variables) -> dict:
    """
    Given new variables and the default variables create a dictionary of the new variables
    :param post_arguments: The new variables to be created
    :param default_variables: The default variables
    :return: The new variables as a dict
    :raises ValueError if any variable values exceed the allowed maximum
    """
    new_script_arguments = {"standard_vars": {}, "advanced_vars": {}}
    for key, value in post_arguments:
        if 'var-' in key:
            if 'var-advanced-' in key:
                name = key.replace('var-advanced-', '').replace('-', ' ')
                dict_key = "advanced_vars"
            elif 'var-standard-' in key:
                name = key.replace('var-standard-', '').replace('-', ' ')
                dict_key = "standard_vars"
            else:
                continue

            if name is not None:
                if len(value) > InstrumentVariable._meta.get_field('value').max_length:
                    raise ValueError(f'Value given in {name} is too long.')

                if name not in default_variables[dict_key]:
                    continue

                new_script_arguments[dict_key][name] = value

    if not new_script_arguments:
        raise ValueError('No variables were found to be submitted.')

    return new_script_arguments


@login_and_uows_valid
@check_permissions
@render_with('configure_new_runs.html')
def configure_new_runs(request, instrument=None, start=0, end=0, experiment_reference=0):
    """
    Handles request to view instrument variables
    """
    instrument_name = instrument

    if request.method == 'POST':
        return configure_new_runs_post(request, instrument_name)
    else:
        return configure_new_runs_get(instrument, start, end, experiment_reference)


def configure_new_runs_post(request, instrument_name):
    """
    Submission to modify variables. Acts on POST request.

    Depending on the parameters it either makes them for a run range (when start is given, end is optional)
    or for experiment reference (when experiment_reference is given).
    """
    start = int(request.POST.get("run_start")) if request.POST.get("run_start", None) else None
    experiment_reference = int(request.POST.get("experiment_reference_number")) if request.POST.get(
        "experiment_reference_number", None) else None

    # [("var-standard-"+name, value) or ("var-advanced-"+name, value)]
    var_list = [t for t in request.POST.items() if t[0].startswith("var-")]

    standard_vars = {
        var[0].replace("var-standard-", ""): var[1]
        for var in var_list if var[0].startswith("var-standard")
    }
    advanced_vars = {
        var[0].replace("var-advanced-", ""): var[1]
        for var in var_list if var[0].startswith("var-advanced")
    }
    all_vars = {"standard_vars": standard_vars, "advanced_vars": advanced_vars}

    reduce_vars = ReductionScript(instrument_name, 'reduce_vars.py')
    reduce_vars_module = reduce_vars.load()
    try:
        args_for_range = InstrumentVariablesUtils.merge_arguments(all_vars, reduce_vars_module)
    except KeyError as err:
        return {
            "message":
            f"Error encountered when processing variable with name: {err}. \
            Please check that the names of the variables in reduce_vars.py match the \
            names of the variables shown in the web app."
        }

    instrument = Instrument.objects.get(name=instrument_name)

    if start and start > 0:
        possible_variables = InstrumentVariable.objects.filter(start_run__lte=start, instrument__name=instrument_name)

        # Makes the variables that will be active for the range START -> END
        # These variables DO NOT track the script - their value will not be updated until:
        # - The user manually sets new variables in the web app
        # - The end run number is passed
        InstrumentVariablesUtils.find_or_make_variables(possible_variables,
                                                        instrument.id,
                                                        args_for_range,
                                                        start,
                                                        from_webapp=True)
    else:
        possible_variables = InstrumentVariable.objects.filter(experiment_reference=experiment_reference,
                                                               instrument__name=instrument_name)
        InstrumentVariablesUtils.find_or_make_variables(possible_variables,
                                                        instrument.id,
                                                        args_for_range,
                                                        start,
                                                        experiment_reference,
                                                        from_webapp=True)
    return redirect('instrument:variables_summary', instrument=instrument_name)


# pylint:disable=too-many-locals
def configure_new_runs_get(instrument_name, start=0, end=0, experiment_reference=0):
    """
    GET for the configure new runs page
    """
    instrument = Instrument.objects.get(name__iexact=instrument_name)

    editing = (start > 0 or experiment_reference > 0)

    last_run = instrument.get_last_for_rerun()

    run_variables = ReductionRunUtils.make_kwargs_from_runvariables(last_run)
    standard_vars = run_variables["standard_vars"]
    advanced_vars = run_variables["advanced_vars"]

    # if a specific start is provided, include vars upcoming for the specific start
    filter_kwargs = {"start_run__gte": start if start else last_run.run_number}
    if end:
        # if an end run is provided - don't show variables outside the [start-end) range
        filter_kwargs["start_run__lt"] = end

    upcoming_variables = instrument.instrumentvariable_set.filter(**filter_kwargs)
    if experiment_reference:
        upcoming_experiment_variables = instrument.instrumentvariable_set.filter(
            experiment_reference=experiment_reference)
    else:
        upcoming_experiment_variables = []

    # Updates the variables values. Experiment variables are chained second
    # so they values will overwrite any changes from the run variables
    for upcoming_var in chain(upcoming_variables, upcoming_experiment_variables):
        name = upcoming_var.name
        if name in standard_vars or not upcoming_var.is_advanced:
            standard_vars[name] = upcoming_var
        elif name in advanced_vars or upcoming_var.is_advanced:
            advanced_vars[name] = upcoming_var

    # Unique, comma-joined list of all start runs belonging to the upcoming variables.
    # This seems to be used to prevent submission if trying to resubmit variables for already
    # configured future run numbers - check the checkForConflicts function
    # This should probably be done by the POST method anyway.. so remove it when
    if upcoming_variables:
        upcoming_run_variables = ','.join({str(var.start_run) for var in upcoming_variables})
    else:
        upcoming_run_variables = ""

    try:
        reduce_vars_variables = VariableUtils.get_default_variables(instrument)
    except (FileNotFoundError, ImportError, SyntaxError) as err:
        return {"message": str(err)}

    current_standard_variables = reduce_vars_variables["standard_vars"]
    current_advanced_variables = reduce_vars_variables["advanced_vars"]
    run_start = start if start else last_run.run_number + 1

    context_dictionary = {
        'instrument': instrument,
        'last_instrument_run': last_run,
        'processing': ReductionRun.objects.filter(instrument=instrument, status=Status.get_processing()),
        'queued': ReductionRun.objects.filter(instrument=instrument, status=Status.get_queued()),
        'standard_variables': standard_vars if standard_vars else current_standard_variables,
        'advanced_variables': advanced_vars if advanced_vars else current_advanced_variables,
        'current_standard_variables': current_standard_variables,
        'current_advanced_variables': current_advanced_variables,
        'run_start': run_start,
        # used to determine whether the current form is for an experiment reference
        'current_experiment_reference': experiment_reference,
        # used to create the link to an experiment reference form, using this number
        'submit_for_experiment_reference': last_run.experiment.reference_number,
        'minimum_run_start': run_start,
        'minimum_run_end': run_start + 1,
        'upcoming_run_variables': upcoming_run_variables,
        'editing': editing,
        'tracks_script': '',
    }

    return context_dictionary



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2021 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #

from selenium.webdriver.remote.webelement import WebElement


class RerunFormMixin:
    @staticmethod
    def _set_field(field, value):
        field.clear()
        new_value = value
        field.send_keys(new_value)

    @property
    def cancel_button(self) -> WebElement:
        """
        Finds and returns the back button for toggling the form on the page.
        """
        return self.driver.find_element_by_id("cancel")

    @property
    def submit_button(self) -> WebElement:
        """
        Finds and returns the back button for toggling the form on the page.
        """
        return self.driver.find_element_by_id("variableSubmit")

    @property
    def variable1_field(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable1")

    @property
    def variable1_field_val(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable1").get_attribute("value")

    @variable1_field.setter
    def variable1_field(self, value) -> None:
        """
        Clears the field and sends the keys to the input field.

        Selenium requires that we clear the field first!
        """
        self._set_field(self.variable1_field, value)

    @property
    def variable_str_field(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_str")

    @property
    def variable_str_field_val(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_str").get_attribute("value")

    @variable_str_field.setter
    def variable_str_field(self, value) -> None:
        """
        Clears the field and sends the keys to the input field.

        Selenium requires that we clear the field first!
        """
        self._set_field(self.variable_str_field, value)

    @property
    def variable_int_field(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_int")

    @property
    def variable_int_field_val(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_int").get_attribute("value")

    @variable_int_field.setter
    def variable_int_field(self, value) -> None:
        """
        Clears the field and sends the keys to the input field.

        Selenium requires that we clear the field first!
        """
        self._set_field(self.variable_int_field, value)

    @property
    def variable_float_field(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_float")

    @property
    def variable_float_field_val(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_float").get_attribute("value")

    @variable_float_field.setter
    def variable_float_field(self, value) -> None:
        """
        Clears the field and sends the keys to the input field.

        Selenium requires that we clear the field first!
        """
        self._set_field(self.variable_float_field, value)

    @property
    def variable_listint_field(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_listint")

    @property
    def variable_listint_field_val(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_listint").get_attribute("value")

    @variable_listint_field.setter
    def variable_listint_field(self, value) -> None:
        """
        Clears the field and sends the keys to the input field.

        Selenium requires that we clear the field first!
        """
        self._set_field(self.variable_listint_field, value)

    @property
    def variable_liststr_field(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_liststr")

    @property
    def variable_liststr_field_val(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_liststr").get_attribute("value")

    @variable_liststr_field.setter
    def variable_liststr_field(self, value) -> None:
        """
        Clears the field and sends the keys to the input field.

        Selenium requires that we clear the field first!
        """
        self._set_field(self.variable_liststr_field, value)

    @property
    def variable_none_field(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_none")

    @property
    def variable_none_field_val(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_none").get_attribute("value")

    @variable_none_field.setter
    def variable_none_field(self, value) -> None:
        """
        Clears the field and sends the keys to the input field.

        Selenium requires that we clear the field first!
        """
        self._set_field(self.variable_none_field, value)

    @property
    def variable_empty_field(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_empty")

    @property
    def variable_empty_field_val(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_empty").get_attribute("value")

    @variable_empty_field.setter
    def variable_empty_field(self, value) -> None:
        """
        Clears the field and sends the keys to the input field.

        Selenium requires that we clear the field first!
        """
        self._set_field(self.variable_empty_field, value)

    @property
    def variable_bool_field(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_bool")

    @property
    def variable_bool_field_val(self) -> WebElement:
        """
        Finds and returns the variabl1 input field
        """
        return self.driver.find_element_by_id("var-standard-variable_bool").get_attribute("value")

    @variable_bool_field.setter
    def variable_bool_field(self, value) -> None:
        """
        Clears the field and sends the keys to the input field.

        Selenium requires that we clear the field first!
        """
        if self.variable_bool_field.is_selected() != value:
            self.driver.execute_script("arguments[0].click();", self.variable_bool_field)



# ############################################################################### #
# Autoreduction Repository : https://github.com/ISISScientificComputing/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Custom manage.py command to add a super user to the database
"""
# pylint:disable=imported-auth-user
from django.contrib.auth.models import User
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """
    Adds a super user to the database
    """
    help = 'Creates a super user in the database'

    def handle(self, *args, **options):
        """
        (THIS SHOULD ONLY BE USED FOR DEVELOPMENT)
        handler to add super user to database
        """
        User.objects.filter(username='super').delete()
        User.objects.create_superuser('super', '', 'super')
